"""This module contains the `Icepak` class."""

from __future__ import absolute_import
import csv
import math
import os
import re
from .application.AnalysisIcepak import FieldAnalysisIcepak
from .desktop import exception_to_desktop
from .generic.general_methods import generate_unique_name, aedt_exception_handler
from .modules.Boundary import BoundaryObject
from  collections import OrderedDict


class Icepak(FieldAnalysisIcepak):
    """Icepak application interface.

    This class allows you to connect to an existing Icepack design or create a
    new Icepak design if one does not exist.

    Parameters
    ----------
    projectname : str, optional
        Name of the project to select or the full path to the project
        or AEDTZ archive to open.  The default is ``None``, in which
        case an attempt is made to get an active project. If no 
        projects are present, an empty project is created.
    designname : str, optional
        Name of the design to select. The default is ``None``, in 
        which case an attempt is made to get an active design. If no
        designs are present, an empty design is created.
    solution_type : str, optional
        Solution type to apply to the design. The default is
        ``None``, which applies the default type.
    setup_name : str, optional
        Name of the setup to use as the nominal. The default is
        ``None``, in which case the active setup is used or 
        nothing is used.
    specified_version: str, optional
        Version of AEDT to use. The default is ``None``, in which case
        the active version or latest installed version is  used.
    NG : bool, optional
        Whether to launch AEDT in the non-graphical mode. The default 
        is ``False``, which launches AEDT in the graphical mode.  
    AlwaysNew : bool, optional
        Whether to launch an instance of AEDT in a new thread, even if
        another instance of the ``specified_version`` is active on the
        machine.  The default is ``True``.
    release_on_exit : bool, optional
        Whether to release AEDT on exit. 
    student_version : bool, optional
        Whether to open the AEDT student version. The default is ``False``.

    Examples
    --------

    Create an instance of `Icepak` and connect to an existing Icepak
    design or create a new Icepak design if one does not exist.

    >>> from pyaedt import Icepak
    >>> icepak = Icepak()

    Create an instance of `Icepak` and link to a project named
    ``IcepakProject``. If this project does not exist, create one with
    this name.

    >>> icepak = Icepak("IcepakProject")
    pyaedt Info: Added design ...

    Create an instance of `Icepak` and link to a design named
    ``IcepakDesign1`` in a project named ``IcepakProject``.

    >>> icepak = Icepak("IcepakProject", "IcepakDesign1")
    pyaedt Info: Added design 'IcepakDesign1' of type Icepak.

    Create an instance of `Icepak` and open the specified project,
    which is ``myfile.aedt``.

    >>> icepak = Icepak("myfile.aedt")
    pyaedt Info: Added design ...

    Create an instance of Icepak using the 2021 R1 release and
    open the specified project, which is ``myfile.aedt``.

    >>> icepak = Icepak(specified_version="2021.1", projectname="myfile.aedt")

    """
    
    def __init__(self, projectname=None, designname=None, solution_type=None, setup_name=None,
                 specified_version=None, NG=False, AlwaysNew=False, release_on_exit=False, student_version=False):
        FieldAnalysisIcepak.__init__(self, "Icepak", projectname, designname, solution_type, setup_name,
                                     specified_version, NG, AlwaysNew, release_on_exit, student_version)

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        """Push exit up to parent object Design."""
        if ex_type:
            exception_to_desktop(self, ex_value, ex_traceback)

    @property
    def existing_analysis_sweeps(self):
        """List the existing analysis setups.       

        Returns
        -------
        list
            List of all defined analysis setup names in the Maxwell design.

        """
        setup_list = self.existing_analysis_setups
        sweep_list=[]
        s_type = self.solution_type
        for el in setup_list:
                sweep_list.append(el + " : " +s_type)
        return sweep_list

    @aedt_exception_handler
    def assign_openings(self, air_faces):
        """Assign openings to a list of faces.

        Parameters
        ----------
        air_faces : list
            List of face names.

        Returns
        -------
        :class:`pyaedt.modules.Boundary.BoundaryObject`
            Boundary object when successful or ``None`` when failed.

        Examples
        --------

        Create an opening boundary for the faces of the "USB_GND" object.

        >>> faces = icepak.modeler.primitives["USB_GND"].faces
        >>> face_names = [face.id for face in faces]
        >>> boundary = icepak.assign_openings(face_names)
        pyaedt Info: Face List boundary_faces created
        pyaedt Info: Opening Assigned

        """
        boundary_name = generate_unique_name("Opening")
        self.modeler.create_face_list(air_faces, "boundary_faces")
        props = {}
        air_faces = self.modeler._convert_list_to_ids(air_faces)

        props["Faces"] = air_faces
        props["Temperature"] = "AmbientTemp"
        props["External Rad. Temperature"] = "AmbientRadTemp"
        props["Inlet Type"] = "Pressure"
        props["Total Pressure"] = "AmbientPressure"
        bound = BoundaryObject(self, boundary_name, props, 'Opening')
        if bound.create():
            self.boundaries.append(bound)
            self._messenger.add_info_message("Opening Assigned")
            return bound
        return None

    @aedt_exception_handler
    def assign_2way_coupling(self,setup_name=None, number_of_iterations=2, continue_ipk_iterations=True, ipk_iterations_per_coupling=20):
        """Assign two-way coupling to a setup.

        Parameters
        ----------
        setup_name : str, optional
            Name of the setup. The default is ``None``.
        number_of_iterations : int, optional
            Number of iterations. The default is ``2``.
        continue_ipk_iterations : bool, optional
           Whether to continue Icepak iterations. The default is ``True``.
        ipk_iterations_per_coupling : int, optional
            Additional iterations per coupling. The default is ``20``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Examples
        --------

        >>> icepak.assign_2way_coupling("Setup1", 1, True, 10)
        True

        """
        if not setup_name:
            if self.setups:
                setup_name =self.setups[0].name
            else:
                self._messenger.add_error_message("No Setup Defined")
                return False
        self.oanalysis_setup.AddTwoWayCoupling(setup_name, ["NAME:Options", "NumCouplingIters:=", number_of_iterations,
                                                            "ContinueIcepakIterations:=", continue_ipk_iterations,
                                                            "IcepakIterationsPerCoupling:=",
                                                            ipk_iterations_per_coupling])
        return True

    @aedt_exception_handler
    def create_source_blocks_from_list(self, list_powers, assign_material=True, default_material="Ceramic_material"):
        """Assign to a box in Icepak the sources that come from the CSV file. 
        
        Assignment is made by name.

        Parameters
        ----------
        list_powers : list
            List of input powers. It is a list of lists. For example, 
            ``[["Obj1", 1], ["Obj2", 3]]``. The list can contain multiple 
            columns for power inputs.
        assign_material : bool, optional
            Whether to assign a material. The default is ``True``.
        default_material : str, optional
            Default material. The default is ``"Ceramic_material"``.

        Returns
        -------
        list
            List of boundaries inserted.

        Examples
        --------

        Create block boundaries from each box in the list. 

        >>> box1 = icepak.modeler.primitives.create_box([1, 1, 1], [3, 3, 3], "BlockBox1", "copper")
        >>> box2 = icepak.modeler.primitives.create_box([2, 2, 2], [4, 4, 4], "BlockBox2", "copper")
        >>> blocks = icepak.create_source_blocks_from_list([["BlockBox1", 2], ["BlockBox2", 4]])
        >>> blocks[1].props
        {'Objects': ['BlockBox1'], 'Block Type': 'Solid', 'Use External Conditions': False, 'Total Power': '2W'}
        >>> blocks[3].props
        {'Objects': ['BlockBox2'], 'Block Type': 'Solid', 'Use External Conditions': False, 'Total Power': '4W'}

        """
        oObjects = self.modeler.primitives.solid_names
        listmcad = []
        num_power = None
        for row in list_powers:
            if not num_power:
                num_power = len(row)-1
                self["P_index"] = 0
            if row[0] in oObjects:
                listmcad.append(row)
                if num_power>1:
                    self[row[0]+"_P"] = str(row[1:])
                    out = self.create_source_block(row[0], row[0]+"_P[P_index]", assign_material, default_material)

                else:
                    out = self.create_source_block(row[0], str(row[1]) + "W", assign_material, default_material)
                if out:
                    listmcad.append(out)

        return listmcad

    @aedt_exception_handler
    def create_source_block(self, object_name, input_power, assign_material=True, material_name="Ceramic_material", use_object_for_name=True):
        """Create a source block for an object.

        Parameters
        ----------
        object_name : str
            Name of the object.
        input_power : str or var
            Input power. 
        assign_material : bool, optional
            Whether to assign a material. The default is ``True``.
        material_name :
            Material to assign if ``assign_material=True``. The default is ``"Ceramic_material"``.
        use_object_for_name : bool, optional
             The default is ``True``.

        Returns
        -------
        :class:`pyaedt.modules.Boundary.BoundaryObject`
            Boundary object when successful or ``None`` when failed.

        Examples
        --------

        >>> box = icepak.modeler.primitives.create_box([5, 5, 5], [1, 2, 3], "BlockBox3", "copper")
        >>> block = icepak.create_source_block("BlockBox3", "1W", False)
        >>> block.props
        {'Objects': ['BlockBox3'], 'Block Type': 'Solid', 'Use External Conditions': False, 'Total Power': '1W'}

        """
        if assign_material:
            self.modeler.primitives[object_name].material_name = material_name
        props={}
        props["Objects"] = [object_name]
        props["Block Type"] = "Solid"
        props["Use External Conditions"] = False
        props["Total Power"] = input_power
        if use_object_for_name:
            boundary_name = object_name
        else:
            boundary_name = generate_unique_name("Block")

        bound = BoundaryObject(self, boundary_name, props, 'Block')
        if bound.create():
            self.boundaries.append(bound)
            return bound
        return None

    @aedt_exception_handler
    def create_source_power(self, face_id, input_power="0W", thermal_condtion="Total Power",
                            surface_heat="0irrad_W_per_m2", temperature="AmbientTemp", 
                            radiate=False, source_name=None):
        """Create a source power for a face.

        Parameters
        ----------
        face_id : int
            Face ID.
        input_power : str, float, or int, optional
            Input power.
        surface_heat : str, optional
            Surface heat.  The default is ``"0irrad_W_per_m2"``.
        temperature : str, optional
            Type of the temperature. The default is ``"AmbientTemp"``.
        radiate : bool
            Whether to enable radiation. The default is ``False``.
        source_name : str, optional
            Name of the source. The default is ``None``.

        Returns
        -------
        :class:`pyaedt.modules.Boundary.BoundaryObject`
            Boundary object when successful or ``None`` when failed.

        Examples
        --------

        Create two source boundaries from one box, one on the top face and one on the bottom face.

        >>> box = icepak.modeler.primitives.create_box([0, 0, 0], [20, 20, 20], name="SourceBox")
        >>> source1 = icepak.create_source_power(box.top_face.id, input_power="2W")
        >>> source1.props["Total Power"]
        '2W'
        >>> source2 = icepak.create_source_power(box.bottom_face.id, thermal_condtion="Fixed Temperature", temperature="28cel")
        >>> source2.props["Temperature"]
        '28cel'

        """
        if not source_name:
            source_name = generate_unique_name("Source")
        props={}
        props["Faces"] = [face_id]
        props["Thermal Condition"] = thermal_condtion
        props["Total Power"] = input_power
        props["Surface Heat"] = surface_heat
        props["Temperature"] = temperature
        props["Radiation"] = OrderedDict({"Radiate" : radiate})
        bound = BoundaryObject(self, source_name, props, 'SourceIcepak')
        if bound.create():
            self.boundaries.append(bound)
            return bound

    @aedt_exception_handler
    def create_network_block(self, object_name, power, rjc, rjb, gravity_dir, top, assign_material=True,
                             default_material="Ceramic_material", use_object_for_name=True):
        """Create a network block.

        Parameters
        ----------
        object_name : str
            Name of the object to create the block for.
        power : str or var
            Input power.
        rjc :
            RJC value.
        rjb :
            RJB value.
        gravity_dir :
            Gravity direction from -X to +Z. Options are ``0`` through ``5``. 
        top :
            Board bounding value in mm of the top face.
        assign_material : bool, optional
            Whether to assign a material. The default is ``True``.
        default_material : str, optional
            Default material if ``assign_material=True``. The default is ``"Ceramic_material"``.
        use_object_for_name : bool, optional
             The default is ``True``.

        Returns
        -------
        :class:`pyaedt.modules.Boundary.BoundaryObject`
            Boundary object.

        Examples
        --------

        >>> box = icepak.modeler.primitives.create_box([4, 5, 6], [5, 5, 5], "NetworkBox1", "copper")
        >>> block = icepak.create_network_block("NetworkBox1", "2W", 20, 10, icepak.GravityDirection.ZNeg, 1.05918)
        >>> block.props["Nodes"]["Internal"][0]
        '2W'
        
        """
        if object_name in self.modeler.primitives.object_names:
            faces = self.modeler.primitives.get_object_faces(object_name)
            k = 0
            faceCenter = {}
            for f in faces:
                faceCenter[f] = self.modeler.oeditor.GetFaceCenter(int(f))
                faceCenter[f] = [round(float(i), 1) for i in faceCenter[f]]
                k = k + 1
            fcmax = -1e9
            fcmin = 1e9
            fcrjc = None
            fcrjb = None
            for fc in faceCenter:
                fc1 = faceCenter[fc]
                if fc1[gravity_dir] < fcmin:
                    fcmin = fc1[gravity_dir]
                    fcrjc = int(fc)
                if fc1[gravity_dir] > fcmax:
                    fcmax = fc1[gravity_dir]
                    fcrjb = int(fc)
            if fcmax < float(top):
                app = fcrjc
                fcrjc = fcrjb
                fcrjb = app
            if assign_material:
                self.modeler.primitives[object_name].material_name = default_material
            props = {}
            if use_object_for_name:
                boundary_name = object_name
            else:
                boundary_name = generate_unique_name("Block")
            props["Faces"] = [fcrjc, fcrjb]
            props["Nodes"] = OrderedDict({"Face" + str(fcrjc): [fcrjc, "NoResistance"],
                              "Face" + str(fcrjb): [fcrjb, "NoResistance"], "Internal": [power]})
            props["Links"] = OrderedDict({"Link1": ["Face" + str(fcrjc), "Internal", "R", str(rjc) + "cel_per_w"],
                              "Link2": ["Face" + str(fcrjb), "Internal", "R", str(rjb) + "cel_per_w"]})
            props["SchematicData"] = OrderedDict({})
            bound = BoundaryObject(self, boundary_name, props, 'Network')
            if bound.create():
                self.boundaries.append(bound)
                self.modeler.primitives[object_name].solve_inside = False
                return bound
            return None

    @aedt_exception_handler
    def create_network_blocks(self, input_list, gravity_dir, top, assign_material=True,
                             default_material="Ceramic_material"):
        """Create network blocks from CSV files.

        Parameters
        ----------
        input_list : list
            List of sources with inputs ``rjc``,  ``rjb``, and ``power``. 
            For example, ``[[Objname1, rjc, rjb, power1, power2, ...], [Objname2, rjc2, rbj2, power1, power2, ...]]``.
        gravity_dir : int
            Gravity direction from -X to +Z. Options are ``0`` to ``5``.
        top :
            Board bounding value in millimeters of the top face.
        assign_material : bool, optional
            Whether to assign a material. The default is ``True``.
        default_material : str, optional
            Default material if ``assign_material=True``. The default is ``"Ceramic_material"``.
       
        Returns
        -------
        list
            List of boundary objects created.

        Examples
        --------

        Create network boundaries from each box in the list. 

        >>> box1 = icepak.modeler.primitives.create_box([1, 2, 3], [10, 10, 10], "NetworkBox2", "copper")
        >>> box2 = icepak.modeler.primitives.create_box([4, 5, 6], [5, 5, 5], "NetworkBox3", "copper")
        >>> blocks = icepak.create_network_blocks([["NetworkBox2", 20, 10, 3], ["NetworkBox3", 4, 10, 2]],
        ...                                        icepak.GravityDirection.ZNeg, 1.05918, False)
        >>> blocks[0].props["Nodes"]["Internal"]
        ['3W']
        
        """
        objs = self.modeler.primitives.solid_names
        countpow = len(input_list[0])-3
        networks = []
        for row in input_list:
            if row[0] in objs:
                if countpow>1:
                    self[row[0] + "_P"] = str(row[3:])
                    self["P_index"] = 0
                    out = self.create_network_block(row[0], row[0] + "_P[P_index]", row[1], row[2], gravity_dir, top,
                                                    assign_material, default_material)
                else:
                    if not row[3]:
                        pow="0W"
                    else:
                        pow=str(row[3])+"W"
                    out = self.create_network_block(row[0], pow, row[1], row[2], gravity_dir,
                                                    top, assign_material, default_material)
                if out:
                    networks.append(out)
        return networks

    @aedt_exception_handler
    def assign_surface_monitor(self, face_name, monitor_type="Temperature", monitor_name=None):
        """Assign a surface monitor.

        Parameters
        ----------
        face_name : str
            Name of the face.
        monitor_type : str, optional
            Type of the monitor.  The default ``"Temperature"``.
        monitor_name : str, optional
            Name of the monitor.  The default is ``None``, in which case
            a default name is assigned automatically.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        if not monitor_name:
            monitor_name = generate_unique_name("Monitor")
        oModule = self.odesign.GetModule("Monitor")
        oModule.AssignFaceMonitor(["NAME:" + monitor_name, "Quantities:=", [monitor_type], "Objects:=", [face_name]])
        return True

    @aedt_exception_handler
    def assign_point_monitor(self, point_position, monitor_type="Temperature", monitor_name=None):
        """Create and assign a point monitor.

        Parameters
        ----------
        point_position : str
            Name of the point.
        monitor_type : str, optional
            Type of the monitor. The default is ``"Temperature"``.
        monitor_name : str, optional
            Name of the monitor. The default is ``None``, in which case 
            a default name is assigned automatically.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        point_name = generate_unique_name("Point")
        self.modeler.oeditor.CreatePoint(
            ["NAME:PointParameters", "PointX:=", self.modeler.primitives._arg_with_dim(point_position[0]), "PointY:=",
             self.modeler.primitives._arg_with_dim(point_position[1]), "PointZ:=",
             self.modeler.primitives._arg_with_dim(point_position[2])],
            ["NAME:Attributes", "Name:=", point_name, "Color:=", "(143 175 143)"])
        if not monitor_name:
            monitor_name = generate_unique_name("Monitor")
        oModule = self.odesign.GetModule("Monitor")
        oModule.AssignPointMonitor(["NAME:" + monitor_name, "Quantities:=", [monitor_type], "Points:=", [point_name]])
        return True


    @aedt_exception_handler
    def assign_block_from_sherlock_file(self, csv_name):
        """Assign block power to components based on a CSV file from Sherlock.

        Parameters
        ----------
        csv_name : str
            Name of the CSV file.

        Returns
        -------
        type
            Total power applied.
        
        """
        with open(csv_name) as csvfile:
            csv_input = csv.reader(csvfile)
            component_header = next(csv_input)
            data = list(csv_input)
            k = 0
            component_data = {}
            for el in component_header:
                component_data[el] = [i[k] for i in data]
                k += 1
        total_power = 0
        i=0
        all_objects = self.modeler.primitives.object_names
        for power in component_data["Applied Power (W)"]:
            try:
                float(power)
                if "COMP_" + component_data["Ref Des"][i] in all_objects:
                    status = self.create_source_block("COMP_" + component_data["Ref Des"][i], str(power)+"W",assign_material=False)
                    if not status:
                        self._messenger.add_warning_message(
                            "Warning. Block {} skipped with {}W power".format(component_data["Ref Des"][i], power))
                    else:
                        total_power += float(power)
                        # print("Block {} created with {}W power".format(component_data["Ref Des"][i], power))
                elif component_data["Ref Des"][i] in all_objects:
                    status = self.create_source_block(component_data["Ref Des"][i], str(power)+"W",assign_material=False)
                    if not status:
                        self._messenger.add_warning_message(
                            "Warning. Block {} skipped with {}W power".format(component_data["Ref Des"][i], power))
                    else:
                        total_power += float(power)
                        # print("Block {} created with {}W power".format(component_data["Ref Des"][i], power))
            except:
                pass
            i += 1
        self._messenger.add_info_message("Blocks inserted with total power {}W".format(total_power))
        return total_power

    @aedt_exception_handler
    def assign_priority_on_intersections(self, component_prefix="COMP_"):
        """Validate an Icepak design.
        
        If there are intersections, priorities are automatically applied to overcome simulation issues.

        Parameters
        ----------
        component_prefix: str, optional
            Component prefix to search for. The default is ``"COMP_"``.

        Returns
        -------
        bool
             ``True`` when successful, ``False`` when failed.     
        
        """
        temp_log = os.path.join(self.project_path, "validation.log")
        validate = self.odesign.ValidateDesign(temp_log)
        self.save_project()
        i = 2
        if validate == 0:
            priority_list = []
            with open(temp_log, 'r') as f:
                lines = f.readlines()
                for line in lines:
                    if '[error]' in line and component_prefix in line and "intersect" in line:
                        id1 = line.find(component_prefix)
                        id2 = line[id1:].find("\"")
                        name = line[id1:id1 + id2]
                        if name not in priority_list:
                            priority_list.append(name)
            print("{} Intersections have been found. Applying Priorities".format(len(priority_list)))
            for objname in priority_list:
                self.mesh.add_priority(1, [objname], priority=i)
                i += 1
        return True

    @aedt_exception_handler
    def find_top(self, gravityDir):
        """Find the top location of the layout given a gravity.

        Parameters
        ----------
        gravityDir :
            Gravity direction from -X to +Z. Choices are ``0`` to ``5``

        Returns
        -------
        float
            Top position
        
        """
        dirs = ["-X", "+X", "-Y", "+Y", "-Z", "+Z"]
        for dir in dirs:
            argsval = ["NAME:" + dir + " Padding Data", "Value:=", "0"]
            args = ["NAME:AllTabs", ["NAME:Geometry3DCmdTab", ["NAME:PropServers", "Region:CreateRegion:1"],
                                     ["NAME:ChangedProps", argsval]]]
            self.modeler.oeditor.ChangeProperty(args)
        oBoundingBox = self.modeler.get_model_bounding_box()
        if gravityDir < 3:
            return oBoundingBox[gravityDir + 3]
        else:
            return oBoundingBox[gravityDir - 3]

    @aedt_exception_handler
    def create_parametric_fin_heat_sink(self, hs_height=100, hs_width=100, hs_basethick=10, pitch=20, thick=10, length=40, height=40,
                                        draftangle=0, patternangle=10, separation=5, symmetric=True, symmetric_separation=20,
                                        numcolumn_perside=2, vertical_separation=10, matname="Al-Extruded", center = [0,0,0],
                                        plane_enum=0, rotation=0, tolerance=1e-3):
        """Create a parametric heat sink.

        Parameters
        ----------
        hs_height : int, optional
            Height of the heat sink. The default is ``100``.
        hs_width: int, optional
            Width of the heat sink. The default is ``100``.
        hs_basethick : int, optional
            Thickness of the heat sink base. The default is ``100``.
        pitch: optional
            Pitch of the heat sink. The default is ``10``.
        thick : optional
            Thickness of the heatsink. The default is ``10``.
        length: optional
            The default is ``40``.
        height : optional
            Height of the heatsink.  The default is ``40``.
        draftangle : optional
            Draft angle.  The default is ``0``.
        patternangle : optional
            Pattern angle.  The default is ``10``.
        separation : optional
            The default is ``5``.
        symmetric : bool, optional
            Whether the heatsink is symmetric.  The default is ``True``.
        symmetric_separation: optional
            The default is ``20``.
        numcolumn_perside : int, optional
            Number of columns per side. The default is ``2``.
        vertical_separation: optional
            The default is ``10``.
        matname : str, optional
            Name of the material. The default is ``Al-Extruded``.
        center : list, optional
           Center of the heatsink.  The default is ``[0, 0, 0]``.
        plane_enum : optional
            The default is ``0``.
        rotation : optional
            The default is ``0``.
        tolerance : optional
            The default is ``1e-3``.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        all_objs = self.modeler.primitives.object_names
        self['FinPitch'] = self.modeler.primitives._arg_with_dim(pitch)
        self['FinThickness'] = self.modeler.primitives._arg_with_dim(thick)
        self['FinLength'] = self.modeler.primitives._arg_with_dim(length)
        self['FinHeight'] = self.modeler.primitives._arg_with_dim(height)
        self['DraftAngle'] = draftangle
        self['PatternAngle'] = patternangle
        self['FinSeparation'] = self.modeler.primitives._arg_with_dim(separation)
        self['VerticalSeparation'] = self.modeler.primitives._arg_with_dim(vertical_separation)
        self['HSHeight'] = self.modeler.primitives._arg_with_dim(hs_height)
        self['HSWidth'] = self.modeler.primitives._arg_with_dim(hs_width)
        self['HSBaseThick'] = self.modeler.primitives._arg_with_dim(hs_basethick)
        if numcolumn_perside>1:
            self['NumColumnsPerSide'] = numcolumn_perside
        if symmetric:
            self['SymSeparation'] = self.modeler.primitives._arg_with_dim(symmetric_separation)
        # ipk['PatternDirection'] = 'Y'
        # ipk['LengthDirection'] = 'X'
        # ipk['HeightDirection'] = 'Z'
        self['Tolerance'] = self.modeler.primitives._arg_with_dim(tolerance)

        #self.modeler.primitives.create_box([0, 0, '-HSBaseThick'], ['HSWidth', 'HSHeight', 'FinHeight+HSBaseThick'], "Outline")
        self.modeler.primitives.create_box(['-HSWidth/200', '-HSHeight/200', '-HSBaseThick'], ['HSWidth*1.01', 'HSHeight*1.01', 'HSBaseThick+Tolerance'], "HSBase", matname)
        Fin_Line = []
        Fin_Line.append(self.Position(0, 0, 0))
        Fin_Line.append(self.Position(0, 'FinThickness', 0))
        Fin_Line.append(self.Position('FinLength', 'FinThickness + FinLength*sin(PatternAngle*3.14/180)', 0))
        Fin_Line.append(self.Position('FinLength', 'FinLength*sin(PatternAngle*3.14/180)', 0))
        Fin_Line.append(self.Position(0, 0, 0))
        self.modeler.primitives.create_polyline(Fin_Line, cover_surface=True, name="Fin")
        Fin_Line2 = []
        Fin_Line2.append(self.Position(0, 'sin(DraftAngle*3.14/180)*FinThickness', 'FinHeight'))
        Fin_Line2.append(self.Position(0, 'FinThickness-sin(DraftAngle*3.14/180)*FinThickness', 'FinHeight'))
        Fin_Line2.append(
            self.Position('FinLength', 'FinThickness + FinLength*sin(PatternAngle*3.14/180)-sin(DraftAngle*3.14/180)*FinThickness',
                         'FinHeight'))
        Fin_Line2.append(
            self.Position('FinLength', 'FinLength*sin(PatternAngle*3.14/180)+sin(DraftAngle*3.14/180)*FinThickness', 'FinHeight'))
        Fin_Line2.append(self.Position(0, 'sin(DraftAngle*3.14/180)*FinThickness', 'FinHeight'))
        self.modeler.primitives.create_polyline(Fin_Line2, cover_surface=True, name="Fin_top")
        self.modeler.connect(["Fin", "Fin_top"])
        self.modeler.primitives["Fin"].material_name = matname
        # self.modeler.thicken_sheet("Fin",'-FinHeight')
        num = int((hs_width / (separation+thick))/(max(1-math.sin(patternangle*3.14/180),0.1)))
        self.modeler.duplicate_along_line("Fin", self.Position(0, 'FinSeparation+FinThickness', 0), num,True)
        self.modeler.duplicate_along_line("Fin", self.Position(0, '-FinSeparation-FinThickness', 0), num/4, True)

        all_names = self.modeler.primitives.object_names
        list = [i for i in all_names if "Fin" in i]
        if numcolumn_perside>0:
            self.modeler.duplicate_along_line(list,
                                             self.Position('FinLength+VerticalSeparation', 'FinLength*sin(PatternAngle*3.14/180)',
                                                          0),
                                             'NumColumnsPerSide', True)

        all_names = self.modeler.primitives.object_names
        list = [i for i in all_names if "Fin" in i]
        self.modeler.split(list, self.CoordinateSystemPlane.ZXPlane, "PositiveOnly")
        all_names = self.modeler.primitives.object_names
        list = [i for i in all_names if "Fin" in i]
        self.modeler.create_coordinate_system(self.Position(0, 'HSHeight', 0), mode="view", view="XY", name="TopRight")

        self.modeler.split(list, self.CoordinateSystemPlane.ZXPlane, "NegativeOnly")

        if symmetric:

            self.modeler.create_coordinate_system(self.Position('(HSWidth-SymSeparation)/2', 0, 0), mode="view",
                                                  view="XY", name="CenterRightSep",reference_cs="TopRight")

            self.modeler.split(list, self.CoordinateSystemPlane.YZPlane, "NegativeOnly")
            self.modeler.create_coordinate_system(self.Position('SymSeparation/2', 0, 0),
                                                  mode="view", view="XY", name="CenterRight",reference_cs="CenterRightSep")
            self.modeler.duplicate_and_mirror(list, self.Position(0, 0, 0), self.Position(1, 0, 0))
            Center_Line = []
            Center_Line.append(self.Position('-SymSeparation', 'Tolerance','-Tolerance'))
            Center_Line.append(self.Position('SymSeparation', 'Tolerance', '-Tolerance'))
            Center_Line.append(self.Position('VerticalSeparation', '-HSHeight-Tolerance', '-Tolerance'))
            Center_Line.append(self.Position('-VerticalSeparation', '-HSHeight-Tolerance', '-Tolerance'))
            Center_Line.append(self.Position('-SymSeparation', 'Tolerance', '-Tolerance'))
            self.modeler.primitives.create_polyline(Center_Line, cover_surface=True, name="Center")
            self.modeler.thicken_sheet("Center", '-FinHeight-2*Tolerance')
            all_names = self.modeler.primitives.object_names
            list = [i for i in all_names if "Fin" in i]
            self.modeler.subtract(list, "Center", False)
        else:
            self.modeler.create_coordinate_system(self.Position('HSWidth', 0, 0), mode="view", view="XY",
                                                  name="BottomRight",reference_cs="TopRight")
            self.modeler.split(list, self.CoordinateSystemPlane.YZPlane, "NegativeOnly")
        all_objs2 = self.modeler.primitives.object_names
        list_to_move=[i for i in all_objs2 if i not in all_objs]
        center[0] -= hs_width/2
        center[1] -= hs_height/2
        center[2] += hs_basethick
        self.modeler.set_working_coordinate_system("Global")
        self.modeler.translate(list_to_move, center)
        if plane_enum == self.CoordinateSystemPlane.XYPlane:
            self.modeler.rotate(list_to_move, self.CoordinateSystemAxis.XAxis, rotation)
        elif plane_enum == self.CoordinateSystemPlane.ZXPlane:
            self.modeler.rotate(list_to_move, self.CoordinateSystemAxis.XAxis, 90)
            self.modeler.rotate(list_to_move, self.CoordinateSystemAxis.YAxis, rotation)
        elif plane_enum == self.CoordinateSystemPlane.YZPlane:
            self.modeler.rotate(list_to_move, self.CoordinateSystemAxis.YAxis, 90)
            self.modeler.rotate(list_to_move, self.CoordinateSystemAxis.ZAxis, rotation)
        self.modeler.unite(list_to_move)
        self.modeler.primitives[list_to_move[0]].name = "HeatSink1"
        return True

    @aedt_exception_handler
    def edit_design_settings(self, gravityDir=0, ambtemp=22, performvalidation=False, CheckLevel="None",
                             defaultfluid="air", defaultsolid="Al-Extruded"):
        """Update the main settings of the design.

        Parameters
        ----------
        gravityDir : int, optional
            Gravity direction from -X to +Z. Options are ``0`` through ``5``. 
            The default is ``0``.
        ambtemp : optional 
            Ambient temperature. The default is ``22``.
        performvalidation : bool, optional
            Whether to perform validation. The default is ``False``.
        CheckLevel : optional
            Level of check to perform during validation. The default is ``None``.
        defaultfluid : str, optional
            Default fluid material. The default is ``"air"``.
        defaultsolid : str, optional
            Default solid material. The default is ``"Al-Extruded"``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        AmbientTemp = str(ambtemp)+"cel"
        #
        # Configure design settings for gravity etc
        IceGravity = ["X", "Y", "Z"]
        GVPos = False
        if int(gravityDir) > 2:
            GVPos = True
        GVA = IceGravity[int(gravityDir) - 3]
        self._odesign.SetDesignSettings(
            ["NAME:Design Settings Data", "Perform Minimal validation:=", performvalidation, "Default Fluid Material:=",
             defaultfluid, "Default Solid Material:=", defaultsolid, "Default Surface Material:=",
             "Steel-oxidised-surface",
             "AmbientTemperature:=", AmbientTemp, "AmbientPressure:=", "0n_per_meter_sq",
             "AmbientRadiationTemperature:=", AmbientTemp,
             "Gravity Vector CS ID:=", 1, "Gravity Vector Axis:=", GVA, "Positive:=", GVPos],
            ["NAME:Model Validation Settings",
             "EntityCheckLevel:=", CheckLevel, "IgnoreUnclassifiedObjects:=", False, "SkipIntersectionChecks:=", False])
        return True

    @aedt_exception_handler
    def assign_em_losses(self, designname="HFSSDesign1", setupname="Setup1", sweepname="LastAdaptive", map_frequency=None,
                         surface_objects=[], source_project_name=None, paramlist=[], object_list=[]):
        """Map EM losses to an Icepak design.

        Parameters
        ----------
        designname : string, optional
            Name of the design of the source mapping. The default is ``"HFSSDesign1"``.
        setupname : str, optional
            Name of the EM setup. The default is ``"Setup1"``.
        sweepname : str, optional
            Name of the EM sweep to use for mapping. The default is ``"LastAdaptive"``.
        map_frequency : optional
            String containing the frequency to map. The default is ``None``. 
            The value must be ``None`` for Eigenmode analysis.
        surface_objects : list, optional
            List of objects in the source that are metals. The default is ``[]``.
        source_project_name : str, optional
            Name of the source project. The default is ``None``, in which case the 
            source from the same project is used.
        paramlist :list, optional
            List of all params in the EM to map. The default is ``[]``.
        object_list : list, optional
            List of objects. The default is ``[]``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        self._messenger.add_info_message("Mapping HFSS EM Lossess")
        oName = self.project_name
        if oName == source_project_name or source_project_name is None:
            projname = "This Project*"
        else:
            projname = source_project_name + ".aedt"
        #
        # Generate a list of model objects from the lists made previously and use to map the HFSS losses into Icepak
        #
        if not object_list:
            allObjects = self.modeler.primitives.object_names
            if "Region" in allObjects:
                allObjects.remove("Region")
        else:
            allObjects = object_list[:]

        surfaces = surface_objects
        if map_frequency:
            intr = [map_frequency]
        else:
            intr = []


        argparam = OrderedDict({})
        for el in self.available_variations.nominal_w_values_dict:
            argparam[el] = self.available_variations.nominal_w_values_dict[el]

        for el in paramlist:
            argparam[el]=el

        props = OrderedDict(
            {"Objects": allObjects, "Project": projname, "projname": "ElectronicsDesktop",
             "Design": designname, "Soln": setupname + " : " + sweepname, "Params": argparam,
             "ForceSourceToSolve": True, "PreservePartnerSoln": True, "PathRelativeTo": "TargetProject"})
        props["Intrinsics"]= intr
        props["SurfaceOnly"]= surfaces

        name = generate_unique_name("EMLoss")
        bound = BoundaryObject(self, name, props, "EMLoss")
        if bound.create():
            self.boundaries.append(bound)
            self._messenger.add_info_message('EM losses mapped from design {}.'.format(designname))
            return bound
        return False


    @aedt_exception_handler
    def eval_surface_quantity_from_field_summary(self, faces_list, quantity_name="HeatTransCoeff", savedir=None,
                                                 filename=None, sweep_name=None, parameter_dict_with_values={}):
        """Export the field surface output.
        
        This method exports one CSV file for the specified variation.

        Parameters
        ----------
        faces_list : list
            List of faces to apply.
        quantity_name : str, optional
            Name of the quantity to export. The default is ``"HeatTransCoeff"``.
        savedir : str, optional
            Directory to save the HTC file to. The default is ``None``.
        filename : str, optional
            Name of the HTC file. The default is ``None``.
        sweep_name : str, optional
            Name of the setup and name of the sweep. For example: ``"IcepakSetup1 : SteatyState"``. The default is ``None``.
        parameter_dict_with_values : dict, optional
            Dictionary of parameters defined for the specific setup with values. The default is ``{}``.

        Returns
        -------
        str
            Name of the file.
        
        """
        name=generate_unique_name(quantity_name)
        self.modeler.create_face_list(faces_list,name )
        if not savedir:
            savedir = self.project_path
        if not filename:
            filename = generate_unique_name(self.project_name+quantity_name)
        if not sweep_name:
            sweep_name=self.nominal_sweep
        self.osolution.EditFieldsSummarySetting(
            [
                "Calculation:=", ["Object", "Surface", name, quantity_name, "", "Default", "AmbientTemp"],
            ])
        string = ""
        for el in parameter_dict_with_values:
            string += el + "=\'" + parameter_dict_with_values[el] + "\' "
        filename = os.path.join(savedir, filename + ".csv")
        self.osolution.ExportFieldsSummary(
                [
                    "SolutionName:=", sweep_name,
                    "DesignVariationKey:=", string,
                    "ExportFileName:=", filename,
                    "IntrinsicValue:=", ""
                ])
        return filename

    def eval_volume_quantity_from_field_summary(self, object_list, quantity_name="HeatTransCoeff",  savedir=None,
                                                 filename=None, sweep_name=None, parameter_dict_with_values={}):
        """Export the field volume output.
        
        This method exports one CSV file for the specified variation.

        Parameters
        ----------
        object_list : list
            List of faces to apply.
        quantity_name : str, optional
            Name of the quantity to export. The default is ``"HeatTransCoeff"``.
        savedir : str, optional
            Directory to save the HTC file to. The default is ``None``.
        filename :  str, optional
            Name of the HTC file. The default is ``None``.
        sweep_name :
            Name of the setup and name of the sweep. For example: ``"IcepakSetup1 : SteatyState"``. The default is ``None``.
        parameter_dict_with_values : dict, optional
            Dictionary of parameters defined for the specific setup with values. The default is ``{}``

        Returns
        -------
        str
           Name of the file.
        
        """
        if not savedir:
            savedir = self.project_path
        if not filename:
            filename = generate_unique_name(self.project_name+quantity_name)
        if not sweep_name:
            sweep_name=self.nominal_sweep
        self.osolution.EditFieldsSummarySetting(
            [
                "Calculation:=", ["Object", "Volume", ",".join(object_list), quantity_name, "", "Default", "AmbientTemp"],
            ])
        string = ""
        for el in parameter_dict_with_values:
            string += el + "=\'" + parameter_dict_with_values[el] + "\' "
        filename = os.path.join(savedir, filename + ".csv")
        self.osolution.ExportFieldsSummary(
                [
                    "SolutionName:=", sweep_name,
                    "DesignVariationKey:=", string,
                    "ExportFileName:=", filename,
                    "IntrinsicValue:=", ""
                ])
        return filename

    @aedt_exception_handler
    def UniteFieldsSummaryReports(self, savedir, proj_icepak):
        """Unite the files created by a fields summary for the variations.

        Parameters
        ----------
        savedir : str
           Directory path for saving the file.           
        proj_icepak : str
            Name of the Icepak project.  

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        newfilename = os.path.join(savedir, proj_icepak + "_HTCAndTemp.csv")
        newfilelines = []
        headerwriten = False
        filetoremove = []
        # first variation
        i = 0
        filename = os.path.join(savedir, proj_icepak + "_HTCAndTemp_var" + str(i) + ".csv")
        # iterate the variations
        while os.path.exists(filename):
            with open(filename, 'r') as f:
                lines = f.readlines()
                variation = lines[1]
                # Searching file content for temp and power
                pattern = re.compile(r"DesignVariation,\$AmbientTemp=('.+?')\s+\$PowerIn=('.+?')")
                m = pattern.match(variation)
                temp = m.group(1)
                power = m.group(2)
                if not headerwriten:
                    # write the new file header
                    newfilelines.append("$AmbientTemp,$PowerIn," + lines[4])
                    newfilelines.append("\n")
                    headerwriten = True

                # add the new lines
                newfilelines.append(temp + ',' + power + ',' + lines[5])
                newfilelines.append(temp + ',' + power + ',' + lines[6])

            # search for next variation
            filetoremove.append(filename)
            i += 1
            filename = os.path.join(savedir, proj_icepak + "_HTCAndTemp_var" + str(i) + ".csv")
        # write the new file
        with open(newfilename, 'w') as f:
            f.writelines(newfilelines)
        # remove the single files variation
        for fr in filetoremove:
            os.remove(fr)
        return True


    def export_summary(self, output_dir=None, solution_name=None, type="Object", geometryType="Volume", quantity="Temperature",
                       variation="", variationlist=[]):
        """Export a fields summary of all objects.

        Parameters
        ----------
        output_dir : str, optional
             Name of directory to export the fields summary to. The default is ``None``.
        solution_name : str, optional
             Name of the solution. The default is ``None``.
        type : string, optional
             The default is ``"Object"``.
        geometryType : str, optional
             Type of the geometry. The default is ``"Volume"``.
        quantity : str, optional
             The default is ``"Temperature"``.
        variation : str, optional
             The default is ``""``.
        variationlist : list, optional
             The default is ``[]``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        all_objs = list(self.modeler.oeditor.GetObjectsInGroup("Solids"))
        all_objs_NonModeled = list(self.modeler.oeditor.GetObjectsInGroup("Non Model"))
        all_objs_model = [item for item in all_objs if item not in all_objs_NonModeled]
        arg = []
        self._messenger.add_info_message("Objects Lists " + str(all_objs_model))
        for el in all_objs_model:
            try:
                self.osolution.EditFieldsSummarySetting(
                    ["Calculation:=", [type, geometryType, el, quantity, "", "Default"]])
                arg.append("Calculation:=")
                arg.append([type, geometryType, el, quantity, "", "Default"])
            except Exception as e:
                self._messenger.add_error_message("Object " + el + " not added")
                self._messenger.add_error_message(str(e))
        if not output_dir:
            output_dir = self.project_path
        self.osolution.EditFieldsSummarySetting(arg)
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        if not solution_name:
            solution_name = self.nominal_sweep
        if variation:
            for l in variationlist:
                self.osolution.ExportFieldsSummary(["SolutionName:=",solution_name ,
                                                          "DesignVariationKey:=", variation + "=\'" + str(l) + "\'",
                                                          "ExportFileName:=",
                                                         os.path.join(output_dir,
                                                                      "IPKsummaryReport" + quantity + "_" + str(
                                                                          l) + ".csv")])
        else:
            self.osolution.ExportFieldsSummary(["SolutionName:=", solution_name,
                                                      "DesignVariationKey:=", "",
                                                      "ExportFileName:=",
                                                      os.path.join(output_dir, "IPKsummaryReport" + quantity + ".csv")])
        return True

    @aedt_exception_handler
    def get_radiation_settings(self, radiation):
        """Retrieve radiation settings.

        Parameters
        ----------
        radiation : str
            One of the following:
            
            * ``"Nothing"``
            * ``"Low"``
            * ``"High"``
            * ``"Both"``   

        Returns
        -------

        """
        if radiation == "Nothing":
            lowSideRad = False
            highSideRad = False
        elif radiation == "Low":
            lowSideRad = True
            highSideRad = False
        elif radiation == "High":
            lowSideRad = False
            highSideRad = True
        elif radiation == "Both":
            lowSideRad = True
            highSideRad = True
        return lowSideRad, highSideRad

    @aedt_exception_handler
    def get_link_data(self, linkData):
        """Retrieve a list of linked data.
                
        Parameters
        ----------
        linkData : list
            List of the data to retrieve for links. Options are:
            
            * Project name, if ``None`` use the active project.
            * Design name
            * HFSS solution name, such as ``"HFSS Setup 1 : Last Adaptive"``
            * Force source design simulation, ``True`` or ``False``.
            * Preserve source design solution, ``True`` or ``False``.

        Returns
        -------
        list
            List containing the requested link data.
        
        """
        if linkData[0] is None:
            project_name = "This Project*"
        else:
            project_name = linkData[0].replace("\\","/")

        designName = linkData[1]
        hfssSolutionName = linkData[2]
        forceSourceSimEnabler = linkData[3]
        preserveSrcResEnabler = linkData[4]

        arg = ["NAME:DefnLink",
            "Project:="	, project_name,
            "Product:="		, "ElectronicsDesktop",
            "Design:="		, designName,
            "Soln:="		, hfssSolutionName,
            ["NAME:Params"],
            "ForceSourceToSolve:="	, forceSourceSimEnabler,
            "PreservePartnerSoln:="	, preserveSrcResEnabler,
            "PathRelativeTo:="	, "TargetProject"]

        return arg

    @aedt_exception_handler
    def create_ipk_3dcomponent_pcb(self, compName, setupLinkInfo, solutionFreq, resolution, PCB_CS="Global",
                                   rad="Nothing", extenttype="Bounding Box", outlinepolygon="", powerin="0W",
                                   custom_x_resolution=None, custom_y_resolution=None):
        """Create a PCB component in Icepak that is linked to an HFSS 3D Layout object.

        Parameters
        ----------
        compName : str
            Name of the new PCB component.
        setupLinkInfo : list
            List of the five elements needed to set up the link: 
            ``[projectname, designname, solution name, forcesimulation (bool), preserve results (bool)``.
        solutionFreq :
            Frequency of the solution if cosimulation is requested.
        resolution : int
            Resolution of the mapping.
        PCB_CS : str, optional
            Coordinate system for the PCB. The default is ``"Global"``.
        rad : str, optional
            Radiating faces. The default is ``"Nothing"``.
        extenttype : str, optional
            Type of the extent. Options are ``"Bounding Box"`` and ``"Polygon"``.  
            The default is ``"Bounding Box"``.
        outlinepolygon : str, optional
            Name of the polygon if ``extentype="Polygon"``. The default is ``""``.
        powerin : str, optional
            Power to dissipate if cosimulation is disabled. The default is ``"0W"``.

        Returns
        -------
        bool
             ``True`` when successful, ``False`` when failed.
        
        """
        lowRad, highRad = self.get_radiation_settings(rad)
        hfssLinkInfo = self.get_link_data(setupLinkInfo)
        compInfo = ["NAME:BasicComponentInfo",
                    "ComponentName:=", compName,
                    "Company:=", "",
                    "Company URL:=", "",
                    "Model Number:=", "",
                    "Help URL:=", "",
                    "Version:=", "1.0",
                    "Notes:=", "",
                    "IconType:=", "PCB"]
        compDefinition = ["NAME:NativeComponentDefinitionProvider",
                          "Type:=", "PCB",
                          "Unit:="	, self.modeler.model_units,
                          "MovePlane:="		, "XY",
                          "Use3DLayoutExtents:="	, False,
                          "ExtentsType:="		, extenttype,
                          "OutlinePolygon:="	, outlinepolygon]

        compDefinition += ["CreateDevices:=", False,
                            "CreateTopSolderballs:=", False,
                            "CreateBottomSolderballs:=", False, ]
        compDefinition += ["Resolution:=", int(resolution),
                           ["NAME:LowSide", "Radiate:=", lowRad],
                           ["NAME:HighSide", "Radiate:=", highRad]]


        if custom_x_resolution and custom_y_resolution:
            compDefinition += ["UseThermalLink:=", solutionFreq!="",
                               "CustomResolution:=", True, "CustomResolutionRow:=", custom_x_resolution,
                               "CustomResolutionCol:=", 600]
        else:
            compDefinition += ["UseThermalLink:=", solutionFreq != "",
                               "CustomResolution:=", False]
        if solutionFreq:
                compDefinition += ["Frequency:=", solutionFreq, hfssLinkInfo]
        else:
            compDefinition += ["Power:=", powerin, hfssLinkInfo]

        compInstancePar = ["NAME:InstanceParameters",
                           "GeometryParameters:="	, "",
                           "MaterialParameters:="	, "",
                           "DesignParameters:="	, ""]

        compData = ["NAME:InsertNativeComponentData",
                    "TargetCS:=", PCB_CS,
                    "SubmodelDefinitionName:=", compName,
                    ["NAME:ComponentPriorityLists"],
                    "NextUniqueID:=", 0,
                    "MoveBackwards:=", False,
                    "DatasetType:=", "ComponentDatasetType",
                    ["NAME:DatasetDefinitions"],
                    compInfo,
                    ["NAME:GeometryDefinitionParameters" ,["NAME:VariableOrders"] ],
                    ["NAME:DesignDefinitionParameters",	["NAME:VariableOrders"]	],
                    ["NAME:MaterialDefinitionParameters", ["NAME:VariableOrders"] ],
                    "MapInstanceParameters:=", "DesignVariable",
                    "UniqueDefinitionIdentifier:=", "a731a7c3-f557-4ab7-ae80-6bf8520c6129",
                    "OriginFilePath:="	, "",
                    "IsLocal:="		, False,
                    "ChecksumString:="	, "",
                    "ChecksumHistory:="	, [],
                    "VersionHistory:="	, [],
                    compDefinition,
                    compInstancePar]

        self.modeler.oeditor.InsertNativeComponent(compData)
        self.modeler.primitives.refresh_all_ids()
        self.materials._load_from_project()
        return True

    @aedt_exception_handler
    def create_pcb_from_3dlayout(self, component_name, project_name, design_name, resolution=2,
                                 extenttype="Bounding Box", outlinepolygon="", close_linked_project_after_import=True,
                                 custom_x_resolution=None, custom_y_resolution=None):
        """Create a PCB component in Icepak that is linked to an HFSS 3D Layout object linking only to the geometry file.
        
        .. note::
           No solution is linked.

        Parameters
        ----------
        component_name : str
            Name of the new PCB component created in Icepak.
        project_name : str
            Name of the oroject or the full path to the project.
        design_name : str
            Name of the design.
        resolution : optional
            Resolution of the mapping. The default is `2`.
        extenttype :
            Type of the extent. Options are ``"Polygon"`` and ``"Bounding Box"``. The default is ``"Bounding Box"``.
        outlinepolygon : str, optional
            Name of the outline polygon if ``extentyype="Polygon"``. The default is ``""``.
        close_linked_project_after_import : bool, optional
            Whether to close the linked AEDT project after the import. The default is ``True``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        if project_name == self.project_name:
            project_name = "This Project*"
        link_data = [project_name, design_name,  "<--EDB Layout Data-->", False, False]
        status = self.create_ipk_3dcomponent_pcb(component_name, link_data, "", resolution, extenttype=extenttype,
                                                 outlinepolygon=outlinepolygon, custom_x_resolution=custom_x_resolution,
                                                 custom_y_resolution=custom_y_resolution)

        if close_linked_project_after_import and ".aedt" in project_name:
                prjname = os.path.splitext(os.path.basename(project_name))[0]
                self.close_project(prjname, saveproject=False)
        self._messenger.add_info_message("PCB Component Correctly created in Icepak")
        return status

    @aedt_exception_handler
    def copyGroupFrom(self, groupName, sourceDesign, sourceProject = None, sourceProjectPath = None):
        """Copy a group from another design.

        Parameters
        ----------
        groupName : str
            Name of the group.
        sourceDesign : str
            Name of the source design.
        sourceProject : str, optional
            Name of the source project. The default is ``None``.
        sourceProjectPath : str, optional
            Path to the source project. The default is ``None``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        oName = self.project_name
        if sourceProject == oName or sourceProject is None:
            oSrcProject = self._desktop.GetActiveProject()
        else:
            self._desktop.OpenProject(sourceProjectPath)
            oSrcProject = self._desktop.SetActiveProject(sourceProject)

        oDesign = oSrcProject.SetActiveDesign(sourceDesign)
        oEditor = oDesign.SetActiveEditor("3D Modeler")
        oEditor.Copy(["NAME:Selections", "Selections:=", groupName])

        self.modeler.oeditor.Paste()
        self.modeler.primitives.refresh_all_ids()
        self.materials._load_from_project()
        return True

    @aedt_exception_handler
    def export3DModel(self, fileName, filePath, fileFormat = ".step", object_list = [], removed_objects = []):
        """Export the 3D model.

        Parameters
        ----------
        fileName : str
            Name of the file.   
        filePath : str
            Path to the file.
        fileFormat : str, optional
             Format of the file. The default is ``".step"``.
        object_list : list, optional
             The default is ``[]``.
        removed_objects : list, optional
             The default is ``[]``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        if not object_list:
            allObjects = self.modeler.primitives.object_names
            if removed_objects:
                for rem in removed_objects:
                    allObjects.remove(rem)
            else:
                if "Region" in allObjects:
                    allObjects.remove("Region")
        else:
            allObjects = object_list[:]

        print(allObjects)

        stringa = ','.join(allObjects)
        arg = ["NAME:ExportParameters",
               "AllowRegionDependentPartSelectionForPMLCreation:=", True,
               "AllowRegionSelectionForPMLCreation:=", True,
               "Selections:=", stringa,
               "File Name:=", str(filePath)+"/"+str(fileName)+str(fileFormat),
               "Major Version:=", -1,
               "Minor Version:=", -1]

        self.modeler.oeditor.Export(arg)
        return True


    @aedt_exception_handler
    def globalMeshSettings(self, meshtype, gap_min_elements="1", noOgrids=False, MLM_en=True, MLM_Type="3D",
                           stairStep_en=False, edge_min_elements="1", object="Region"):
        """Create a custom mesh tailored on a PCB design.

        Parameters
        ----------
        meshtype : 
            Type of mesh. Options are ``1``, ``2``, and ``3``, which represent 
            respectively a coarse, standard, or very accurate mesh.
        gap_min_elements : str, optional
            The default is ``"1"``.
        noOgrids : bool, optional
            The default is ``False``.
        MLM_en: bool, optional
            The default is ``True``.
        MLM_Type: str, optional
            The default is ``"3D"``.
        stairStep_en : bool, optional
            The default is ``False``.
        edge_min_elements: str, optional
            The default is ``"1"``.
        object : str, optional
            The default is ``"Region"``.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        oModule = self.odesign.GetModule("MeshRegion")

        oBoundingBox = self.modeler.oeditor.GetModelBoundingBox()
        xsize = abs(float(oBoundingBox[0]) - float(oBoundingBox[3])) / (15 * meshtype * meshtype)
        ysize = abs(float(oBoundingBox[1]) - float(oBoundingBox[4])) / (15 * meshtype * meshtype)
        zsize = abs(float(oBoundingBox[2]) - float(oBoundingBox[5])) / (10 * meshtype)
        MaxSizeRatio = (1 + (meshtype / 2))

        oModule.EditGlobalMeshRegion(
            [
                "NAME:Settings",
                "MeshMethod:="	, "MesherHD",
                "UserSpecifiedSettings:=", True,
                "ComputeGap:="		, True,
                "MaxElementSizeX:="	, str(xsize) + self.modeler.model_units,
                "MaxElementSizeY:="	, str(ysize) + self.modeler.model_units,
                "MaxElementSizeZ:="	, str(zsize) + self.modeler.model_units,
                "MinElementsInGap:="	, gap_min_elements,
                "MinElementsOnEdge:="	, edge_min_elements,
                "MaxSizeRatio:="	, str(MaxSizeRatio),
                "NoOGrids:="		, noOgrids,
                "EnableMLM:="		, MLM_en,
                "EnforeMLMType:="	, MLM_Type,
                "MaxLevels:="		, "0",
                "BufferLayers:="	, "0",
                "UniformMeshParametersType:=", "Average",
                "StairStepMeshing:="	, stairStep_en,
                "MinGapX:="		, str(xsize / 10) + self.modeler.model_units,
                "MinGapY:="		, str(xsize / 10) + self.modeler.model_units,
                "MinGapZ:="		, str(xsize / 10) + self.modeler.model_units,
                "Objects:="		, [object]
            ])
        return True

    @aedt_exception_handler
    def create_meshregion_component(self, scale_factor=1.0, name="Component_Region",
                                          restore_padding_values=[50, 50, 50, 50, 50, 50]):
        """Create a bounding box to be used as Mesh Region in Icepak.

        Parameters
        ----------
        scale_factor : float, optional
             The default is ``1.0``.
        name : str, optional
            The default is "Component_Region"
        restore_padding_values : list, optional
             The default is ``[50,50,50,50,50,50]``.  

        Returns
        -------
        tuple
            Tuple containing the ``(x, y, z)`` distances of the region.
        
        """
        self.modeler.edit_region_dimensions([0, 0, 0, 0, 0, 0])

        verticesID = self.modeler.oeditor.GetVertexIDsFromObject("Region")

        x_values = []
        y_values = []
        z_values = []

        for id in verticesID:
            tmp = self.modeler.oeditor.GetVertexPosition(id)
            x_values.append(tmp[0])
            y_values.append(tmp[1])
            z_values.append(tmp[2])

        scale_factor = scale_factor - 1
        delta_x = (float(max(x_values)) - float(min(x_values))) * scale_factor
        x_max = float(max(x_values)) + delta_x / 2.
        x_min = float(min(x_values)) - delta_x / 2.

        delta_y = (float(max(y_values)) - float(min(y_values))) * scale_factor
        y_max = float(max(y_values)) + delta_y / 2.
        y_min = float(min(y_values)) - delta_y / 2.

        delta_z = (float(max(z_values)) - float(min(z_values))) * scale_factor
        z_max = float(max(z_values)) + delta_z / 2.
        z_min = float(min(z_values)) - delta_z / 2.

        dis_x = str(float(x_max) - float(x_min))
        dis_y = str(float(y_max) - float(y_min))
        dis_z = str(float(z_max) - float(z_min))

        min_position = self.modeler.Position(str(x_min)+"mm", str(y_min)+"mm", str(z_min)+"mm")
        mesh_box = self.modeler.primitives.create_box(min_position, [dis_x+"mm", dis_y+"mm", dis_z+"mm"],
                                                      name)

        self.modeler.primitives[name].model = False

        self.modeler.edit_region_dimensions(restore_padding_values)
        return dis_x, dis_y, dis_z

    @aedt_exception_handler
    def create_temp_point_monitor(self, point_name, point_coord=[0, 0, 0]):
        """Create a temperature monitor for the simulation.

        Parameters
        ----------
        point_name : str
            Name of the temperature monitor.
        point_coord : list, optional
            Point coordinates for the temperature monitor. The default is ``[0, 0, 0]``.
        
        Returns
        -------
        bool
             ``True`` when successful, ``False`` when failed.
        
        """
        arg1 = ["NAME:PointParameters",
               "PointX:="		, point_coord[0],
               "PointY:="		, point_coord[1],
               "PointZ:="		, point_coord[2]]

        arg2 = ["NAME:Attributes",
                "Name:="		, point_name,
                "Color:="		, "(143 175 143)"]

        self.modeler.oeditor.CreatePoint(arg1, arg2)
        monitor = self._odesign.GetModule("Monitor")

        arg = ["NAME:"+str(point_name),
               "Quantities:="		, ["Temperature"],
               "Points:="		, [str(point_name)]]

        monitor.AssignPointMonitor(arg)
        return True

    @aedt_exception_handler
    def delete_em_losses(self, bound_name):
        """Delete the EM losses boundary.

        Parameters
        ----------
        bound_name : str
            Name of the boundary.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        self.oboundary.DeleteBoundaries([bound_name])
        return True

    @aedt_exception_handler
    def delete_pcb_component(self, comp_name):
        """Delete a PCB component.

        Parameters
        ----------
        comp_name : str
            Name of the PCB component.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        arg = ["NAME:Selections",
               "Selections:="		, comp_name]

        self.modeler.oeditor.Delete(arg)
        return True



"""
This module contains these classes: ``BoundaryCommon`` and ``BoundaryObject``.
"""
from collections import OrderedDict
from ..generic.general_methods import aedt_exception_handler
from ..application.DataHandlers import dict2arg
from ..modeler.Object3d import EdgePrimitive, FacePrimitive, VertexPrimitive
class BoundaryCommon(object):
    """ """
    @aedt_exception_handler
    def _get_args(self, props=None):
        """Retrieve boundary properties.
        
        Parameters
        ----------
        props : dict, optional
             The default is ``None``.

        Returns
        -------
        dict
            Dictionary of boundary properties.

        """
        if not props:
            props = self.props
        arg = ["NAME:" + self.name]
        dict2arg(props, arg)
        return arg

    @aedt_exception_handler
    def delete(self):
        """Delete boundaries.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        self._parent.oboundary.DeleteBoundaries([self.name])
        for el in self._parent.boundaries:
            if el.name == self.name:
                self._parent.boundaries.remove(el)
        return True

class BoundaryObject(BoundaryCommon, object):
    """Boundary data and execution class.
    
    Parameters
    ----------
    parent:
    
    name :
    
    props :
    
    boundarytype :
    
        
    """
    def __init__(self, parent, name, props, boundarytype):
        self._parent = parent
        self.name = name
        self.props = props
        self.type = boundarytype

    @aedt_exception_handler
    def _get_args(self, props=None):
        """Retrieve arguments.

        Parameters
        ----------
        props :
            The default is ``None``.

        Returns
        -------
        type
            
        """
        if props is None:
            props = self.props
        arg = ["NAME:" + self.name]
        dict2arg(props, arg)
        return arg

    @aedt_exception_handler
    def create(self):
        """Create a boundary.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if self.type == "PerfectE":
            self._parent.oboundary.AssignPerfectE(self._get_args())
        elif self.type == "PerfectH":
            self._parent.oboundary.AssignPerfectH(self._get_args())
        elif self.type == "Aperture":
            self._parent.oboundary.AssignAperture(self._get_args())
        elif self.type == "Radiation":
            self._parent.oboundary.AssignRadiation(self._get_args())
        elif self.type == "FiniteCond":
            self._parent.oboundary.AssignFiniteCond(self._get_args())
        elif self.type == "LumpedRLC":
            self._parent.oboundary.AssignLumpedRLC(self._get_args())
        elif self.type == "Impedance":
            self._parent.oboundary.AssignImpedance(self._get_args())
        elif self.type == "Anisotropic Impedance":
            self._parent.oboundary.AssignAssignAnisotropicImpedance(self._get_args())
        elif self.type == "Primary":
            self._parent.oboundary.AssignPrimary(self._get_args())
        elif self.type == "Secondary":
            self._parent.oboundary.AssignSecondary(self._get_args())
        elif self.type == "Lattice Pair":
            self._parent.oboundary.AssignLatticePair(self._get_args())
        elif self.type == "HalfSpace":
            self._parent.oboundary.AssignHalfSpace(self._get_args())
        elif self.type == "Multipaction SEE":
            self._parent.oboundary.AssignMultipactionSEE(self._get_args())
        elif self.type == "Fresnel":
            self._parent.oboundary.AssignFresnel(self._get_args())
        elif self.type == "Symmetry":
            self._parent.oboundary.AssignSymmetry(self._get_args())
        elif self.type == 'Zero Tangential H Field':
            self._parent.oboundary.AssignZeroTangentialHField(self._get_args())
        elif self.type == 'Zero Integrated Tangential H Field':
            self._parent.oboundary.AssignIntegratedZeroTangentialHField(self._get_args())
        elif self.type == 'Tangential H Field':
            self._parent.oboundary.AssignTangentialHField(self._get_args())
        elif self.type == 'Insulating':
            self._parent.oboundary.AssignInsulating(self._get_args())
        elif self.type == 'Independent':
            self._parent.oboundary.AssignIndependent(self._get_args())
        elif self.type == 'Dependent':
            self._parent.oboundary.AssignDependent(self._get_args())
        elif self.type == 'InfiniteGround':
            self._parent.oboundary.AssignInfiniteGround(self._get_args())
        elif self.type == 'ThinConductor':
            self._parent.oboundary.AssignThinConductor(self._get_args())
        elif self.type == 'Stationary Wall':
            self._parent.oboundary.AssignStationaryWallBoundary(self._get_args())
        elif self.type == 'Simmetry Wall':
            self._parent.oboundary.AssignSymmetryWallBoundary(self._get_args())
        elif self.type == 'Resistance':
            self._parent.oboundary.AssignResistanceBoundary(self._get_args())
        elif self.type == 'Conducting Plate':
            self._parent.oboundary.AssignConductingPlateBoundary(self._get_args())
        elif self.type == 'Adiabatic Plate':
            self._parent.oboundary.AssignAdiabaticPlateBoundary(self._get_args())
        elif self.type == 'Network':
            self._parent.oboundary.AssignNetworkBoundary(self._get_args())
        elif self.type == 'Grille':
            self._parent.oboundary.AssignGrilleBoundary(self._get_args())
        elif self.type == 'Block':
            self._parent.oboundary.AssignBlockBoundary(self._get_args())
        elif self.type == 'SourceIcepak':
            self._parent.oboundary.AssignSourceBoundary(self._get_args())
        elif self.type == 'Opening':
            self._parent.oboundary.AssignOpeningBoundary(self._get_args())
        elif self.type == 'EMLoss':
            self._parent.oboundary.AssignEMLoss(self._get_args())
        elif self.type == 'ThermalCondition':
            self._parent.oboundary.AssignThermalCondition(self._get_args())
        elif self.type == 'Convection':
            self._parent.oboundary.AssignConvection(self._get_args())
        elif self.type == 'Temperature':
            self._parent.oboundary.AssignTemperature(self._get_args())
        elif self.type == 'RotatingFluid':
            self._parent.oboundary.AssignRotatingFluid(self._get_args())
        elif self.type == 'Frictionless':
            self._parent.oboundary.AssignFrictionlessSupport(self._get_args())
        elif self.type == 'FixedSupport':
            self._parent.oboundary.AssignFixedSupport(self._get_args())
        elif self.type == "Voltage":
            self._parent.oboundary.AssignVoltage(self._get_args())
        elif self.type == "VoltageDrop":
            self._parent.oboundary.AssignVoltageDrop(self._get_args())
        elif self.type == "Current":
            self._parent.oboundary.AssignCurrent(self._get_args())
        elif self.type == 'Balloon':
            self._parent.oboundary.AssignBalloon(self._get_args())
        elif self.type == "Winding" or self.type == "Winding Group":
            self._parent.oboundary.AssignWindingGroup(self._get_args())
        elif self.type == "VectorPotential":
            self._parent.oboundary.AssignVectorPotential(self._get_args())
        elif self.type == "CoilTerminal":
            self._parent.oboundary.AssignCoilTerminal(self._get_args())
        elif self.type == "Coil":
            self._parent.oboundary.AssignCoil(self._get_args())
        elif self.type == 'Source':
            self._parent.oboundary.AssignSource(self._get_args())
        elif self.type == 'Sink':
            self._parent.oboundary.AssignSink(self._get_args())
        elif self.type == 'CircuitPort':
            self._parent.oboundary.AssignCircuitPort(self._get_args())
        elif self.type == 'LumpedPort':
            self._parent.oboundary.AssignLumpedPort(self._get_args())
        elif self.type == 'WavePort':
            self._parent.oboundary.AssignWavePort(self._get_args())
        elif self.type == 'AutoIdentify':
            self._parent.oboundary.AutoIdentifyPorts(["NAME:Faces", self.props["Faces"]],self.props["IsWavePort"],
                   ["NAME:ReferenceConductors", self.props["ReferenceConductors"]], self.name,
                   self.props["RenormalizeModes"])
        else:
            return False
        return True

    @aedt_exception_handler
    def update(self):
        """Update the boundary.
                
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if self.type == "PerfectE":
            self._parent.oboundary.EditPerfectE(self.name, self._get_args())
        elif self.type == "PerfectH":
            self._parent.oboundary.EditPerfectH(self.name, self._get_args())
        elif self.type == "Aperture":
            self._parent.oboundary.EditAperture(self.name, self._get_args())
        elif self.type == "Radiation":
            self._parent.oboundary.EditRadiation(self.name, self._get_args())
        elif self.type == "FiniteCond":
            self._parent.oboundary.EditFiniteCond(self.name, self._get_args())
        elif self.type == "LumpedRLC":
            self._parent.oboundary.EditLumpedRLC(self.name, self._get_args())
        elif self.type == "Impedance":
            self._parent.oboundary.EditImpedance(self.name, self._get_args())
        elif self.type == "Anisotropic Impedance":
            self._parent.oboundary.EditAssignAnisotropicImpedance(self.name, self._get_args())
        elif self.type == "Primary":
            self._parent.oboundary.EditPrimary(self.name, self._get_args())
        elif self.type == "Secondary":
            self._parent.oboundary.EditSecondary(self.name, self._get_args())
        elif self.type == "Lattice Pair":
            self._parent.oboundary.EditLatticePair(self.name, self._get_args())
        elif self.type == "HalfSpace":
            self._parent.oboundary.EditHalfSpace(self.name, self._get_args())
        elif self.type == "Multipaction SEE":
            self._parent.oboundary.EditMultipactionSEE(self.name, self._get_args())
        elif self.type == "Fresnel":
            self._parent.oboundary.EditFresnel(self.name, self._get_args())
        elif self.type == "Symmetry":
            self._parent.oboundary.EditSymmetry(self.name, self._get_args())
        elif self.type == 'Zero Tangential H Field':
            self._parent.oboundary.EditZeroTangentialHField(self.name, self._get_args())
        elif self.type == 'Zero Integrated Tangential H Field':
            self._parent.oboundary.EditIntegratedZeroTangentialHField(self.name, self._get_args())
        elif self.type == 'Tangential H Field':
            self._parent.oboundary.EditTangentialHField(self.name, self._get_args())
        elif self.type == 'Insulating':
            self._parent.oboundary.EditInsulating(self.name, self._get_args())
        elif self.type == 'Independent':
            self._parent.oboundary.EditIndependent(self.name, self._get_args())
        elif self.type == 'Dependent':
            self._parent.oboundary.EditDependent(self.name, self._get_args())
        elif self.type == 'InfiniteGround':
            self._parent.oboundary.EditInfiniteGround(self.name, self._get_args())
        elif self.type == 'ThinConductor':
            self._parent.oboundary.EditThinConductor(self.name, self._get_args())
        elif self.type == 'Stationary Wall':
            self._parent.oboundary.EditStationaryWallBoundary(self.name, self._get_args())
        elif self.type == 'Simmetry Wall':
            self._parent.oboundary.EditSymmetryWallBoundary(self.name, self._get_args())
        elif self.type == 'Resistance':
            self._parent.oboundary.EditResistanceBoundary(self.name, self._get_args())
        elif self.type == 'Conducting Plate':
            self._parent.oboundary.EditConductingPlateBoundary(self.name, self._get_args())
        elif self.type == 'Adiabatic Plate':
            self._parent.oboundary.EditAdiabaticPlateBoundary(self.name, self._get_args())
        elif self.type == 'Network':
            self._parent.oboundary.EditNetworkBoundary(self.name, self._get_args())
        elif self.type == 'Grille':
            self._parent.oboundary.EditGrilleBoundary(self.name, self._get_args())
        elif self.type == 'Opening':
            self._parent.oboundary.EditOpeningBoundary(self.name, self._get_args())
        elif self.type == 'EMLoss':
            self._parent.oboundary.EditEMLoss(self.name, self._get_args())
        elif self.type == 'Block':
            self._parent.oboundary.EditBlockBoundary(self.name,self._get_args())
        elif self.type == 'SourceIcepak':
            self._parent.oboundary.EditSourceBoundary(self._get_args())
        elif self.type == "Voltage":
            self._parent.oboundary.EditVoltage(self.name, self._get_args())
        elif self.type == "VoltageDrop":
            self._parent.oboundary.EditVoltageDrop(self.name, self._get_args())
        elif self.type == "Current":
            self._parent.oboundary.Current(self.name, self._get_args())
        elif self.type == "Winding" or self.type == "Winding Group":
            self._parent.oboundary.EditWindingGroup(self.name, self._get_args())
        elif self.type == "VectorPotential":
            self._parent.oboundary.EditVectorPotential(self.name, self._get_args())
        elif self.type == "CoilTerminal":
            self._parent.oboundary.EditCoilTerminal(self.name, self._get_args())
        elif self.type == "Coil":
            self._parent.oboundary.EditCoil(self.name, self._get_args())
        elif self.type == 'Source':
            self._parent.oboundary.EditSource(self.name, self._get_args())
        elif self.type == 'Sink':
            self._parent.oboundary.EditSink(self.name, self._get_args())
        elif self.type == 'CircuitPort':
            self._parent.oboundary.EditCircuitPort(self.name, self._get_args())
        elif self.type == 'LumpedPort':
            self._parent.oboundary.EditLumpedPort(self.name, self._get_args())
        elif self.type == 'WavePort':
            self._parent.oboundary.EditWavePort(self.name, self._get_args())
        else:
            return False
        return True

    @aedt_exception_handler
    def update_assignment(self):
        """Update the assignment.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if "Faces" in self.props:
            faces = self.props["Faces"]
            faces_out = []
            if type(faces) is not list:
                faces = [faces]
            for f in faces:
                if type(f) is EdgePrimitive or type(f) is FacePrimitive or type(f) is VertexPrimitive:
                    faces_out.append(f.id)
                else:
                    faces_out.append(f)
            self._parent.oboundary.ReassignBoundary(["Name:"+self.name, "Faces:=", faces_out])
        elif "Objects" in self.props:

            self._parent.oboundary.ReassignBoundary(["Name:"+self.name, "Objects:=", self.props["Objects"]])
        else:
            return False
        return True
    



import os
# Import exception handling here due to:
# https://github.com/pyansys/PyAEDT/pull/243

try:
    from .generic.general_methods import aedt_exception_handler, generate_unique_name, retry_ntimes
    from .hfss3dlayout import Hfss3dLayout
    from .hfss import Hfss
    from .circuit import Circuit
    from .q3d import Q2d, Q3d
    from .siwave import Siwave
    from .icepak import Icepak
    from .edb import Edb
    from .maxwell import Maxwell2d, Maxwell3d
    from .mechanical import Mechanical
    from .rmxprt import Rmxprt
    from .simplorer import Simplorer
    from .desktop import Desktop
    from .emit import Emit
except:
    from .generic.general_methods import aedt_exception_handler, generate_unique_name, retry_ntimes
    from .hfss3dlayout import Hfss3dLayout
    from .hfss import Hfss
    from .circuit import Circuit
    from .q3d import Q2d, Q3d
    from .siwave import Siwave
    from .icepak import Icepak
    from .edb import Edb
    from .maxwell import Maxwell2d, Maxwell3d
    from .mechanical import Mechanical
    from .rmxprt import Rmxprt
    from .simplorer import Simplorer
    from .desktop import Desktop
    from .emit import Emit



from distutils.core import setup

from _setup_common import name, version, author, maintainer, maintainer_email, description, long_description, packages, data_files, license, classifiers

setup(
    name=name,
    version=version,
    author=author,
    maintainer=maintainer,
    maintainer_email=maintainer_email,
    description=description,
    long_description=long_description,
    packages=packages,
    data_files=data_files,
    license=license,
    classifiers=classifiers,
)


"""
This module contains these classes: `AEDTMessageManager`, `EDBMessageManager`, 
and `MessageList`.

This module provides all functionalities for logging errors and messages
in both AEDT and the log file.
"""

import sys
import logging
import os
from ..generic.general_methods import aedt_exception_handler, generate_unique_name
message_levels = {'Global': 0,
                  'Project': 1,
                  'Design': 2}

class Msg():
    (INFO, WARNING, ERROR, FATAL) = range(4)


class MessageList:
    """Collects and returns messages from the AEDT Message Manager for a specified project name and design name.

    Parameters
    ---------
    msg_list : list 
        List of messages extracted from AEDT by the :class:`AEDTMessageManager` class.
    project_name : str
        Name of the project. The default is the active project.
    design_name : str 
        Name of the design within the specified project. The default is the active design.


    Attributes
    ----------
    global_level : list of str
        List of strings representing the message content at the global level of the message manager

    project_level : list of str
        List of strings representing the message content of the specifiec project

    design_level : list of str
        List of strings representing the message content for the specified design within the specified project

    """
    def __init__(self, msg_list, project_name, design_name):
        self.global_level = []
        self.project_level = []
        self.design_level = []
        global_label = "Project: *Global - Messages, "
        project_label = "Project: {}, ".format(project_name)
        design_label = "Project: {}, Design: {} (".format(project_name, design_name)
        for line in msg_list:
            # Find the first instance of '[' to get the message context
            loc = line.find('[')
            if loc < 0:
                # Format is not clear - append to global
                self.global_level.append(line)
            else:
                line_label = line[0:loc]
                msg_txt = line[loc:].strip()
                if design_label in line_label:
                    self.design_level.append(msg_txt)
                elif project_label in line_label:
                    self.project_level.append(msg_txt)
                elif global_label in line_label:
                    self.global_level.append(msg_txt)


class AEDTMessageManager(object):
    """Manages AEDT messaging to both the logger and the AEDT Message Manager window.
  
    Parameters
    ----------
    parent :
        The default is ``None``.

    Attributes
    ----------
    messages : list
       List of messages.

    Methods
    -------
    add_error message
    add_warning_message
    add_info_message
    add debug_message (deprecated)
    add message
    get_design_messages
    clear_messages

    Examples
    --------
    Log the three types of messages.

    >>> from pyaedt.hfss import Hfss
    >>> hfss = Hfss()
    >>> hfss._messenger.add_info_message("This is an info message on a design", "Design")
    >>> hfss._messenger.add_warning_message("This is a global warning message", "Global")
    >>> hfss._messenger.add_error_message("This is a project error message", "Project")

    """
    def __init__(self, parent=None):
        self._parent = parent
        if not parent:
            if 'oDesktop' in dir(sys.modules['__main__']):
                self.MainModule = sys.modules['__main__']
                self._desktop = self.MainModule.oDesktop
                self._log_on_desktop = os.getenv('PYAEDT_DESKTOP_LOGS', 'True').lower() in ('true', '1', 't', 'yes', 'y')
            else:
                self._log_on_desktop = False
                self._desktop = None
        else:
            self._desktop = self._parent._desktop
            self._log_on_desktop = os.getenv('PYAEDT_DESKTOP_LOGS', 'True').lower() in ('true', '1', 't')
        self._log_on_file = os.getenv('PYAEDT_FILE_LOGS', 'True').lower() in ('true', '1', 't')
        self._log_on_screen = os.getenv('PYAEDT_SCREEN_LOGS', 'True').lower() in ('true', '1', 't')

        if self._log_on_file:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = None

    @property
    def messages(self):
        """Message manager content for the active project and design.

        Returns
        -------
        list
           List of messages for the active project and design.
           
        """
        return self.get_messages(self._project_name, self._design_name)

    @aedt_exception_handler
    def get_messages(self, project_name, design_name):
        """Retrieve the Message Manager content for a specified project and design.
        
        If the specified project and design names are invalid, they are ignored.
        
        Parameters
        ----------
        project_name : str
            Name of the project.
        design_name : str
            Name of the design within the specified project.     

        Returns
        -------
        list
            List of messages for the specified project and design.

        """
        if self._log_on_desktop:
            global_message_data = self._desktop.GetMessages("", "", 0)
            message_data = MessageList(global_message_data, project_name, design_name)
            return message_data

    @aedt_exception_handler
    def add_error_message(self, message_text, level=None):
        """Add a type 2 "Error" message to the Message Manager tree.

        Also add an error message to the logger if the handler is present.

        Parameters
        ----------
        message_text : str
            Text to display as the error message.
        level : str, optional
            Level to add the error message to. Options are ``"Global"``, 
            ``"Project"``, and ``"Design"``. The default is ``None``, 
            in which case the error message gets added to the ``"Design"``
            level.

        Examples
        --------
        Add an error message to the AEDT Message Manager.

        >>> hfss._messenger.add_error_message("Project Error Message", "Project")

        """
        self.add_message(2, message_text, level)

    @aedt_exception_handler
    def add_warning_message(self, message_text, level=None):
        """Add a type 1 "Warning" message to the Message Manager tree.
        
        Also add a warning message to the logger if the handler is present.

        Parameters
        ----------
        message_text : str
            Text to display as the warning message.
        level : str, optional
            Level to add the warning message to. Options are ``"Global"``, 
            ``"Project"``, and ``"Design"``. The default is ``None``, 
            in which case the warning message gets added to the ``"Design"``
            level.

        Examples
        --------
        Add a warning message to the AEDT Message Manager.

        >>> hfss._messenger.add_warning_message("Project warning message")

        """
        self.add_message(1, message_text, level)

    @aedt_exception_handler
    def add_info_message(self, message_text, level=None):
        """Add a type 0 "Info" message to the active design level of the Message Manager tree.

        Also add an info message to the logger if the handler is present.

        Parameters
        ----------
        message_text : str
            Text to display as the info message.
        level : str, optional
            Level to add the info message to. Options are ``"Global"``, 
            ``"Project"``, and ``"Design"``. The default is ``None``, 
            in which case the info message gets added to the ``"Design"``
            level.

        Examples
        --------
        Add an info message at the global level.

        >>> hfss._messenger.add_info_message("Global warning message", "Global")

        """
        self.add_message(0, message_text, level)

    @aedt_exception_handler
    def add_debug_message(self, message_text, level=None):
        """Deprecated in favor of :func:`MessageList.add_info_message`.
        .. deprecated:: 0.2.0
           Use the method :func:`MessageList.add_info_message`.
        """
        self.add_info_message(message_text, level)

    @aedt_exception_handler
    def add_message(self, type, message_text, level=None, proj_name=None, des_name=None):
        """Pass a parameterized message to the Message Manager to specify the type and project or design level.

        Parameters
        ----------
        type : int
            Type of the message. Options are:

            * 0: Info
            * 1: Warning
            * 2: Error
        
        message_text : str
            Text to display as the message.
        level : str, optional
            Level to add the message to. Options are ``"Global"``, 
            ``"Project"``, and ``"Design"``. The default is ``None``, 
            in which case the message gets added to the ``"Design"``level.
        proj_name : str, optional
            Name of the project.
        des_name : str, optional
            Name of the design.

        """
        if not proj_name:
            proj_name = ""

        if not des_name:
            des_name = ""

        if not level:
            level = 'Design'

        assert level in message_levels, "Message level must be `Design', 'Project', or 'Global'."

        if self._log_on_desktop:
            if not proj_name and message_levels[level] > 0:
                proj_name = self._project_name
            if not des_name and message_levels[level] > 1:
                des_name = self._design_name
            if des_name and ";" in des_name:
                des_name = des_name[des_name.find(";")+1:]
            self._desktop.AddMessage(proj_name, des_name, type, message_text)

        if len(message_text) > 250:
            message_text = message_text[:250] + "..."

        # Print to stdout and to logger

        if self._log_on_screen:
            if type == 0:
                print("pyaedt Info: {}".format(message_text))
            elif type == 1:
                print("pyaedt Warning: {}".format(message_text))
            elif type == 2:
                print("pyaedt Error: {}".format(message_text))
        if self._log_on_file:
            if type == 0 and self.logger:
                self.logger.info(message_text)
            elif type == 1 and self.logger:
                self.logger.warning(message_text)
            elif type == 2 and self.logger:
                self.logger.error(message_text)

    @aedt_exception_handler
    def clear_messages(self, proj_name=None, des_name=None, level=2):
        """Clear all messages.

        Parameters
        ----------
        proj_name : str, optional
            Name of project. The default is ``None``, in which case messages
            are cleared for the current project. If blank, messages are cleared
            for all projects.
        des_name : str, optional
            Name of the design within the specified project. The default
            is ``None,`` in which case the current design is used. 
            If blank, all designs are used.
        level : int, optional
            Level of the messages to clear. Options are:

            * ``0`` : Clear all info messages.
            * ``1`` : Clear all info and warning messages.
            * ``2`` : Clear all info, warning, and error messages.
            * ``3`` : Clear all messages, which include info, warning,
              error, and fatal-error messages.
              
            The default is ``2.``

        Examples
        --------
        Clear all messages in the current design and project.

        >>> hfss.clear_messages(level=3)

        """
        if self._log_on_desktop:
            if not proj_name:
                proj_name = self._project_name
            if not des_name:
                des_name = self._design_name
            self._desktop.ClearMessages(proj_name, des_name, level)

    @property
    def _oproject(self):
        return self._desktop.GetActiveProject()

    @property
    def _odesign(self):
        return self._oproject.GetActiveDesign()

    @property
    def _design_name(self):
        try:
            return self._odesign.GetName()
        except AttributeError:
            return ""

    @property
    def _project_name(self):
        try:
            return self._oproject.GetName()
        except AttributeError:
            return ""


class EDBMessageManager(object):
    """Manages EDB messaging to the logger.
    
    Parameters
    ----------
    project_dir : str, optional
        The default is ``None``.
    """

    def __init__(self, project_dir=None):
        self._log_on_file = os.getenv('PYAEDT_FILE_LOGS', 'True').lower() in ('true', '1', 't')
        self._log_on_screen = os.getenv('PYAEDT_SCREEN_LOGS', 'True').lower() in ('true', '1', 't')
        if self._log_on_file:
            self.logger = logging.getLogger(__name__)
            if not project_dir:
                if os.name == "posix":
                    project_dir = "/tmp"
                else:
                    project_dir = "C:\\Temp"
            if not self.logger.handlers:
                logging.basicConfig(
                    filename=os.path.join(project_dir, "EDBTLib.log"),
                    level=logging.DEBUG,
                    format='%(asctime)s:%(name)s:%(levelname)-8s:%(message)s',
                    datefmt='%Y/%m/%d %H.%M.%S')
                self.logger = logging.getLogger(__name__)
        else:
            self.logger = None

    def add_error_message(self, message_text):
        """Add a type 2 "Error" message to the logger.

        Parameters
        ----------
        message_text : str
            Text to display as the message.
                
        """
        if self._log_on_file:
            self.logger.error(message_text)
        if self._log_on_screen:
            print(message_text)


    def add_warning_message(self, message_text):
        """Add a "Warning" message to the logger.

        message_text : str
            Text to display as the message.
        
        """
        if self._log_on_file:
            self.logger.warning(message_text)
        if self._log_on_screen:
            print(message_text)

    def add_info_message(self, message_text):
        """Add an "Info" message to the logger.

        message_text : str
            Text to display as the message.

        """
        if self._log_on_file:
            self.logger.info(message_text)
        if self._log_on_screen:
            print(message_text)



import warnings
import random
import string
from collections import OrderedDict
from decimal import Decimal
from ..generic.general_methods import aedt_exception_handler, generate_unique_name
from ..modeler.Object3d import EdgePrimitive, FacePrimitive, VertexPrimitive
try:
    import clr
    clr.AddReference("System.Collections")
    from System.Collections.Generic import List
    clr.AddReference("System")
    from System import Double, Array
except ImportError:
    warnings.warn("Pythonnet is needed to run pyaedt")

@aedt_exception_handler
def tuple2dict(t, d):
    """

    Parameters
    ----------
    t :
        
    d :
        

    Returns
    -------

    """
    k = t[0]
    v = t[1]
    if type(v) is list and len(t) > 2:
        d[k] = v
    elif type(v) is list and len(t) == 2 and not v:
        d[k] = None
    elif type(v) is list and type(v[0]) is tuple and len(
            t) == 2:  # len check is to avoid expanding the list with a 3rd element=None
        d[k] = OrderedDict()
        for tt in v:
            tuple2dict(tt, d[k])
    else:
        d[k] = v


@aedt_exception_handler
def dict2arg(d, arg_out):
    """

    Parameters
    ----------
    d :
        
    arg_out :
        

    Returns
    -------

    """
    for k, v in d.items():
        if type(v) is OrderedDict or type(v) is dict:
            arg = ["NAME:" + k]
            dict2arg(v, arg)
            arg_out.append(arg)
        elif v is None:
            arg_out.append(["NAME:" + k])
        elif type(v) is list and len(v)>0 and (type(v[0]) is OrderedDict or type(v[0]) is dict):
            for el in v:
                arg = ["NAME:" + k]
                dict2arg(el, arg)
                arg_out.append(arg)

        else:
            arg_out.append(k + ":=")
            if type(v) is EdgePrimitive or type(v) is FacePrimitive or type(v) is VertexPrimitive:
                arg_out.append(v.id)
            else:
                arg_out.append(v)


@aedt_exception_handler
def arg2dict(arg, dict_out):
    """

    Parameters
    ----------
    arg :
        
    dict_out :
        

    Returns
    -------

    """
    if arg[0] == "NAME:DimUnits" or arg[0] == "NAME:Points":
        dict_out[arg[0][5:]] = list(arg[1:])
    elif arg[0][:5] == 'NAME:':
        top_key = arg[0][5:]
        dict_in = OrderedDict()
        i = 1
        while i < len(arg):
            if (type(arg[i]) is list or type(arg[i]) is tuple) and arg[i][0][:5] == 'NAME:':
                arg2dict(arg[i], dict_in)
                i += 1
            elif arg[i][-2:] == ':=':
                dict_in[arg[i][:-2]] = arg[i + 1]

                i += 2
            else:
                raise ValueError('Incorrect data argument format')
        dict_out[top_key] = dict_in
    else:
        raise ValueError('Incorrect data argument format')



@aedt_exception_handler
def create_list_for_csharp(input_list, return_strings=False):
    """

    Parameters
    ----------
    input_list :
        
    return_strings :
         (Default value = False)

    Returns
    -------

    """
    if return_strings:
        col=List[str]()
    else:
        col=List[Double]()

    for el in input_list:
        if return_strings:
            col.Add(str(el))
        else:
            col.Add(el)
    return col


@aedt_exception_handler
def create_table_for_csharp(input_list_of_list, return_strings=True):
    """

    Parameters
    ----------
    input_list_of_list :
        
    return_strings :
         (Default value = True)

    Returns
    -------

    """
    new_table = List[List[str]]()
    for col in input_list_of_list:
        newcol=create_list_for_csharp(col, return_strings)
        new_table.Add(newcol)
    return new_table


@aedt_exception_handler
def format_decimals(el):
    """

    Parameters
    ----------
    el :
        

    Returns
    -------

    """
    if float(el) > 1000:
        num = "{:,.0f}".format(Decimal(el))
    elif float(el) > 1:
        num = "{:,.3f}".format(Decimal(el))
    else:
        num = "{:.3E}".format(Decimal(el))
    return num


@aedt_exception_handler
def random_string(length=6):
    """Generate a random string

    Parameters
    ----------
    length :
        length of the random string (Default value = 6)

    Returns
    -------
    type
        random string

    """
    char_set = string.ascii_uppercase + string.digits
    random_str = ''.join(random.sample(char_set, int(length)))
    return random_str



class SolutionType(object):
    """SolutionType class.
    
    The class provides the name of the default solution type.
    """
    class Hfss(object):
        """Hfss class."""
        (DrivenModal, DrivenTerminal, EigenMode, Transient, SBR) = (
            "DrivenModal", "DrivenTerminal", "EigenMode", "Transient Network", "SBR+")

    class Maxwell3d(object):
        """Maxwell3d class."""
        (Transient, Magnetostatic, EddyCurrent, ElectroStatic, ElectroDCConduction, ElectroDCTransient) = \
            ("Transient", "Magnetostatic", "EddyCurrent", "Electrostatic", "ElectroDCConduction",
             "ElectricTransient")

    class Maxwell2d(object):
        """Maxwell 2D class."""
        (TransientXY, TransientZ, MagnetostaticXY, MagnetostaticZ, EddyCurrentXY, EddyCurrentZ, ElectroStaticXY,
         ElectroStaticZ, ElectroDCConductionX, ElectroDCConductionZ, ElectroDCTransientXY, ElectroDCTransientZ) = \
            ("TransientXY", "TransientZ", "MagnetostaticXY", "MagnetostaticZ", "EddyCurrentXY", "EddyCurrentZ",
             "ElectrostaticXY", "ElectrostaticZ", "ElectroDCConductionXY","ElectroDCConductionZ",
             "ElectricTransientXY", "ElectricTransientZ")

    class Icepak(object):
        """Icepak class."""
        (SteadyTemperatureAndFlow, SteadyTemperatureOnly, SteadyFlowOnly, TransientTemperatureAndFlow,
         TransientTemperatureOnly, TransientFlowOnly,) = (
            "SteadyTemperatureAndFlow", "SteadyTemperatureOnly", "SteadyFlowOnly", "TransientTemperatureAndFlow",
            "TransientTemperatureOnly", "TransientFlowOnly")

    class Circuit(object):
        """Circuit class."""
        (NexximLNA, NexximDC, NexximTransient, NexximQuickEye, NexximVerifEye, NexximAMI, NexximOscillatorRSF,
         NexximOscillator1T, NexximOscillatorNT, NexximHarmonicBalance1T, NexximHarmonicBalanceNT, NexximSystem,
         NexximTVNoise, HSPICE, TR) = ("NexximLNA", "NexximDC", "NexximTransient",
                                       "NexximQuickEye", "NexximVerifEye", "NexximAMI", "NexximOscillatorRSF",
                                       "NexximOscillator1T",
                                       "NexximOscillatorNT", "NexximHarmonicBalance1T", "NexximHarmonicBalanceNT",
                                       "NexximSystem", "NexximTVNoise", "HSPICE", "TR")

    class Mechanical(object):
        """Mechanical class."""
        (Thermal, Structural, Modal) = ("Thermal", "Structural", "Modal")
class SetupTypes(object):
    """SetupTypes class.
     This class provides constants for the default setup types."""
    (HFSSDrivenAuto, HFSSDrivenDefault, HFSSEigen, HFSSTransient, HFSSSBR, MaxwellTransient, Magnetostatic, EddyCurrent,
     Electrostatic, ElectrostaticDC, ElectricTransient, SteadyTemperatureAndFlow, SteadyTemperatureOnly, SteadyFlowOnly,
     Matrix, NexximLNA, NexximDC, NexximTransient, NexximQuickEye, NexximVerifEye, NexximAMI, NexximOscillatorRSF,
     NexximOscillator1T, NexximOscillatorNT, NexximHarmonicBalance1T, NexximHarmonicBalanceNT, NexximSystem,
     NexximTVNoise, HSPICE, HFSS3DLayout, Open, Close, MechTerm, MechModal, GRM, TR, TransientTemperatureAndFlow,
     TransientTemperatureOnly, TransientFlowOnly,) = range(0, 39)







class BLNDChamferRangeLaw(object):
    """BLNDChamferRangeLaw class."""
    (BLNDConstantRange,
     BLNDVariableRange) = range(0, 2)


class BLNDFilletRadiusLaw(object):
    """BLNDFilletRadiusLaw class."""
    (BLNDConstantRadius,
     BLNDVariableRadius) = range(0, 2)


class BLNDFilletType(object):
    """BLNDFilletType class."""
    (BLNDRound,
     BLNDMitered) = range(0, 2)


class CoordinateSystemAxis(object):
    """CoordinateSystemAxis class."""
    (XAxis, YAxis, ZAxis) = range(0, 3)


class CoordinateSystemPlane(object):
    """CoordinateSystemPlane class."""
    (XYPlane, YZPlane, ZXPlane) = range(0, 3)


class Plane(object):
    """Plane class."""
    (XYPlane, YZPlane, ZXPlane, ISO) = ("XY", "YZ", "ZX", "iso")


class GravityDirection(object):
    """GravityDirection class."""
    (XNeg, YNeg, ZNeg, XPos, YPos, ZPos) = range(0, 6)


class MessageSeverity(object):
    """MessageSeverity class."""
    (InfoMessage,
     WarningMessage,
     ErrorMessage,
     FatalMessage,
     IncompleteMessage) = range(0, 5)


class ParamDataType(object):
    """ParamDataType class."""
    (Unknown,
     Int,
     Double,
     String,
     Bool,
     Position) = range(-1, 5)


class ParamPropFlag(object):
    """ParamPropFlag class."""
    (Unknown,
     NoFlag,
     ReadOnly,
     MustBeInt) = range(-1, 3)
    MustBeReal = 4
    Hidden = 8


class ParamPropType(object):
    """ParamPropTyp class."""
    (Unknown,
     Text,
     Menu,
     Number,
     Value,
     FileName,
     Checkbox,
     Position) = range(-1, 7)


class PartPropertyFlags(object):
    """PartPropertyFlags class."""
    PropNonModel = 1
    PropDisplayWireFrame = 2
    PropReadOnly = 4
    PostprocessingGeometry = 8
    PropInvisible = 16
    PropShowDirection = 32
    PropDummy = 33


class PolylineSegmentType(object):
    """PolylineSegmentType class."""
    (LineSegment,
     ArcSegment,
     SplineSegment,
     AngularArcSegment) = range(0, 4)


class PolylineXSegmentType(object):
    """PolylineXSegmentType class."""
    (NONE,
     Line,
     Circle,
     Rectangle,
     Trapezoid) = range(0, 5)


class SplitWhichSideToKeep(object):
    """SplitWhichSideToKeep class."""
    (SplitKeepBoth,
     SplitKeepPositiveOnly,
     SplitKeepNegativeOnly) = range(0, 3)


class SweepDraftType(object):
    """SweepDraftType class."""
    (ExtendedDraft,
     RoundDraft,
     NaturalDraft,
     MixedDraft) = range(0, 4)


class UnitType(object):
    """UnitType  class."""
    (NoUnit,
     LengthUnit,
     AngleUnit) = range(0, 3)



"""
This module contains these classes: `Mesh` and `Mesh3DOperation`.

This module provides all functionalities for creating and editing the mesh in the 3D tools.

"""
from __future__ import absolute_import

from ..generic.general_methods import aedt_exception_handler, generate_unique_name, MethodNotSupportedError
from ..application.DataHandlers import dict2arg
from collections import OrderedDict, defaultdict

class Mesh3DOperation(object):
    """Mesh3DOperation class.
    
    Parameters
    ----------
    parent :
    
    hfss_setup_name : str
        Name of the HFSS setup.
    name :
    
    props : dict
        Dictionary of the properites. 
    
    """

    def __init__(self, parent, hfss_setup_name, name, props):
        self._parent = parent
        self.name = name
        self.props = props
        self.hfss_setup_name = hfss_setup_name

    @aedt_exception_handler
    def _get_args(self, props=None):
        """Retrieve arguments.

        Parameters
        ----------
        props : dict
            Dictionary of the properties. The default is ``None``, in which
            case the default properties are retrieved.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
            
        """
        if not props:
            props = self.props
        arg = ["NAME:" + self.name]
        dict2arg(props, arg)
        return arg

    @aedt_exception_handler
    def create(self):
        """Create a mesh.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        self._parent.omeshmodule.AddMeshOperation(self.hfss_setup_name, self._get_args())
        return True

    @aedt_exception_handler
    def update(self):
        """Update the mesh.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        self._parent.omeshmodule.EditMeshOperation(self.hfss_setup_name, self.name, self._get_args())
        return True

    @aedt_exception_handler
    def delete(self):
        """Delete the mesh.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
            
        """
        self._parent.omeshmodule.DeleteMeshOperation(self.hfss_setup_name, self.name,)

        return True



class Mesh(object):
    """Mesh class.
    
    This class provides the main AEDT mesh functionaility. The inherited class
    `AEDTConfig` contains all `_desktop` hierarchical calls needed by this class.
   
    Parameters
    ----------
    parent :

    """

    def __init__(self, parent):
        self._parent = parent
        self.id = 0
        self.meshoperations = self._get_design_mesh_operations()


        pass

    @property
    def omeshmodule(self):
        """Mesh module.
        
        Returns
        -------
        type
            Mesh module object.
        """
        
        return self.odesign.GetModule("SolveSetups")

    @property
    def _messenger(self):
        """_messenger."""
        return self._parent._messenger

    @property
    def odesign(self):
        """Design."""
        return self._parent._odesign

    @property
    def modeler(self):
        """Modeler."""
        return self._parent._modeler

    @aedt_exception_handler
    def delete_mesh_operations(self, setup_name, mesh_name):
        """Remove mesh operations from a setup.
       
        Parameters
        ----------
        setup_name : str
            Name of the setup.
        mesh_name :str
            Name of the mesh.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
            
        """
        for el in self.meshoperations:
            if el.hfss_setup_name == setup_name and el.name == mesh_name:
                el.delete()
                self.meshoperations.remove(el)

        return True

    @aedt_exception_handler
    def _get_design_mesh_operations(self):
        """Retrieve design mesh operations.
        
        Returns
        -------
        type
            Mesh operations object.
        
        """
        meshops = []
        try:
            for ds in self._parent.design_properties['Setup']['Data']:
                if 'MeshOps' in self._parent.design_properties['Setup']['Data'][ds]:
                    for ops in self._parent.design_properties['Setup']['Data'][ds]['MeshOps']:
                        props = self._parent.design_properties['Setup']['Data'][ds]['MeshOps'][ops]
                        meshops.append(Mesh3DOperation(self, ds, ops, props))
        except:
            pass
        return meshops

    @aedt_exception_handler
    def assign_length_mesh(self, setupname, layer_name, net_name, isinside=True, maxlength=1, maxel=1000, meshop_name=None):
        """Assign mesh length.

        Parameters
        ----------
        setupname : str
            Name of the HFSS setup to apply.
        layer_name : str
           Name of the layer.   
        net_name : str
           Name of the net.
        isinside : bool, optional
            Whether the mesh length is inside the selection. The default is ``True``.
        maxlength : float, optional
            Maximum length of the element. The default is ``1`` When ``None``, this 
            parameter is disabled. 
        maxel : int, optional
            Maximum number of elements. The default is ``1000``. When ``None``, this 
            parameter is disabled. 
        meshop_name : str, optional
            Name of the mesh operation. The default is ``None``.
            
        Returns
        -------
        type
            Mesh operation object.
        
        """
        if meshop_name:
            for el in self.meshoperations:
                if el.name == meshop_name:
                    meshop_name = generate_unique_name(meshop_name)
        else:
            meshop_name = generate_unique_name("Length")

        if maxlength is None:
            restrictlength = False
        else:
            restrictlength = True
        length = self.modeler.modeler_variable(maxlength)

        if maxel is None:
            restrictel = False
            numel = "1000"
        else:
            restrictel = True
            numel = str(maxel)
        if maxlength is None and maxel is None:
            self._messenger.add_error_message("mesh not assigned due to incorrect settings")
            return
        if type(layer_name) is list and type(net_name) is list:
            assignment = OrderedDict({"MeshEntityInfo":[]})
            for l, n in zip(layer_name,net_name):
                meshbody = OrderedDict({"Id": -1, "Nam": "", "Layer": l, "Net": n, "OrigNet": n})
                assignment["MeshEntityInfo"].append(OrderedDict({"IsFcSel": False, "EntID": -1, "FcIDs": [], "MeshBody": meshbody, "BBox": []}))
        else:
            meshbody = OrderedDict({"Id": -1, "Nam": "", "Layer": layer_name, "Net": net_name, "OrigNet": net_name})
            assignment = OrderedDict({"MeshEntityInfo": OrderedDict(
                {"IsFcSel": False, "EntID": -1, "FcIDs": [], "MeshBody": meshbody, "BBox": []})})
        props = OrderedDict({"Type": "LengthBased", "RefineInside": isinside, "Enabled": True, "Assignment": assignment,
                             "Region":"", "RestrictElem": restrictel, "NumMaxElem": numel, "RestrictLength": restrictlength,
                             "MaxLength": length})

        mop = Mesh3DOperation(self,setupname,meshop_name, props)
        mop.create()
        self.meshoperations.append(mop)
        return mop


    @aedt_exception_handler
    def assign_skin_depth(self, setupname, layer_name, net_name, skindepth=1, maxelements=None, triangulation_max_length=0.1, numlayers="2",
                          meshop_name=None):
        """Assign skin depth to the mesh.

        Parameters
        ----------
        setupname : str
            Name of the setup.
        layer_name : str
            Name of the layer.
        net_name : str
            Name of the net.
        skindepth : int, optional
            Depth of the skin. The default is ``1``.
        maxelements : float, optional
            Maximum element length. The default is ``None``, which disables this parameter.
        triangulation_max_length : float, optional
            Maximum surface triangulation length. The default is ``0.1``.
        numlayers : str, optional
            Number of layers. The default is ``"2"``.
        meshop_name : str, optional
             Name of the mesh operation. The default is ``None``.

        Returns
        -------
        type
            Mesh operation object.
        
        """
        if meshop_name:
            for el in self.meshoperations:
                if el.name == meshop_name:
                    meshop_name = generate_unique_name(meshop_name)
        else:
            meshop_name = generate_unique_name("SkinDepth")

        if maxelements is None:
            restrictlength = False
            maxelements = "1000"
        else:
            restrictlength = True
        skindepth = self.modeler.modeler_variable(skindepth)
        triangulation_max_length= self.modeler.modeler_variable(triangulation_max_length)
        meshbody = OrderedDict({"Id": -1, "Nam": "", "Layer": layer_name,  "Net": net_name, "OrigNet": net_name})
        assignment = OrderedDict({"MeshEntityInfo": OrderedDict({"IsFcSel": False, "EntID": -1, "FcIDs": [], "MeshBody": meshbody, "BBox": []})})
        props = OrderedDict({"Type": "SkinDepthLengthBased", "Enabled": True, "Assignment": assignment,
                             "Region":"", "SkinDepth": skindepth, "SurfTriMaxLength": triangulation_max_length, "NumLayers": numlayers,
                             "RestrictElem": restrictlength, "NumMaxElem": maxelements})

        mop = Mesh3DOperation(self, setupname, meshop_name, props)
        mop.create()
        self.meshoperations.append(mop)
        return mop



"""
This module contains the `Desktop` class.

This module is used to initialize AEDT and Message Manager to manage AEDT.

You can initialize this module before launching an app or 
have the app automatically initialize it to the latest installed AEDT version.
"""
from __future__ import absolute_import

import os
import sys
import traceback
import logging
import pkgutil
import getpass
import re
from pyaedt.application.MessageManager import AEDTMessageManager
from pyaedt.misc import list_installed_ansysem

pathname = os.path.dirname(__file__)
if os.path.exists(os.path.join(pathname,'version.txt')):
    with open(os.path.join(pathname,'version.txt'), "r") as f:
        pyaedtversion = f.readline()
elif os.path.exists(os.path.join(pathname, "..", 'version.txt')):
    with open(os.path.join(pathname, "..", 'version.txt'), "r") as f:
        pyaedtversion = f.readline()
else:
    pyaedtversion = "X"

_pythonver = sys.version_info[0]

if os.name == 'nt':
    IsWindows = True
else:
    IsWindows = False
logger = logging.getLogger(__name__)

if "IronPython" in sys.version or ".NETFramework" in sys.version:
    import clr  # IronPython C:\Program Files\AnsysEM\AnsysEM19.4\Win64\common\IronPython\ipy64.exe
    _com = 'ironpython'
elif IsWindows:
    import pythoncom
    modules = [tup[1] for tup in pkgutil.iter_modules()]
    if 'clr' in modules:
        import clr
        import win32com.client
        _com = 'pythonnet_v3'
    elif 'win32com' in modules:
        import win32com.client
        _com = 'pywin32'
    else:
        raise Exception("Error. No win32com.client or Pythonnet modules found. Please install them")


def exception_to_desktop(self, ex_value, tb_data):
    """Writes the trace stack to the desktop when a Python error occurs.
    
    The message is added to the AEDT global Message Manager and to the log file (if present).

    Parameters
    ----------
    ex_value : str
        Type of exception.
    tb_data : str
        Traceback information.

    """
    desktop = sys.modules['__main__'].oDesktop
    try:
        oproject = desktop.GetActiveProject()
        proj_name = oproject.GetName()
        try:
            des_name = oproject.GetActiveDesign().GetName()
            if ";" in des_name:
                des_name = des_name.split(";")[1]
        except:
            des_name = ''
    except:
        proj_name = ''
        des_name = ''
    tb_trace = traceback.format_tb(tb_data)
    tblist = tb_trace[0].split('\n')
    desktop.AddMessage(proj_name, des_name, 2, str(ex_value))
    for el in tblist:
        desktop.AddMessage(proj_name, des_name, 2, el)


def update_aedt_registry(key, value, desktop_version="211"):
    """Update the AEDT registry key.
    
    .. note::
       This method is only supported on Windows.
    
    Parameters
    ----------
    key : str
        Registry key.
    value : str
        Value for the registry key. The value includes "" if needed.
    desktop_version : str, optional
        Version of AEDT to use. The default is ``"211"`` 
        to use 2021 R1.
    
    Examples
    --------
    Update the HPC license type for HFSS in the AEDT registry.
    
    >>> update_aedt_registry("HFSS/HPCLicenseType", "12") # doctest: +SKIP
    
    Update the HPC license type for Icepak in the AEDT registry.
    
    >>> update_aedt_registry("Icepak/HPCLicenseType", "8") # doctest: +SKIP
    
    Update the legacy HPC license type for HFSS in the AEDT registry.
    
    >>> update_aedt_registry("HFSS/UseLegacyElectronicsHPC", "0") # doctest: +SKIP
    
    Update the MPI vendor for HFSS in the AEDT registry.
    
    >>> update_aedt_registry("HFSS/MPIVendor", "Intel") # doctest: +SKIP

    """
    if os.name == 'posix':
        import subprocessdotnet as subprocess
    else:
        import subprocess
    desktop_install_dir = os.environ["ANSYSEM_ROOT" + str(desktop_version)]

    with open(os.path.join(desktop_install_dir, "config", "ProductList.txt")) as file:
        product_version = next(file).rstrip()  # get first line

    options = '-set -ProductName {} + product_version -RegistryKey "{}" -RegistryValue "{}"'.format(product_version,
                                                                                                    key, value)
    command = '"{}/UpdateRegistry" {}'.format(desktop_install_dir, options)

    subprocess.call([command])


def release_desktop(close_projects=True, close_desktop=True):
    """Release the AEDT API.

    Parameters
    ----------
    close_projects : bool, optional
        Whether to close the projects opened in the session. The default is ``True``.
    close_desktop : bool, optional
        Whether to close the active AEDT session. The default is ``True``.

    Returns
    -------
    bool
        ``True`` when successful, ``False`` when failed.

    """
    if _com == "pythonnet":
        Module = sys.modules['__main__']
        desktop = Module.oDesktop
        scopeID = desktop.ScopeID
        i = 0
        if close_projects:
            proj_list = desktop.GetProjectList()
            for prj in proj_list:
                desktop.CloseProject(prj)

        while i <= scopeID:
            Module.COMUtil.ReleaseCOMObjectScope(Module.COMUtil.PInvokeProxyAPI, i)
            i += 1
        try:
            del Module.oDesktop
            return True
        except:
            Module.oMessenger.add_info_message("Attributes not present")
            return False

    elif _com == "pythonnet_v3":
        Module = sys.modules['__main__']
        desktop = Module.oDesktop
        i = 0
        if close_projects:
            proj_list = desktop.GetProjectList()
            for prj in proj_list:
                desktop.CloseProject(prj)

        if close_desktop:
            scopeID = 5
            while i <= scopeID:
                Module.COMUtil.ReleaseCOMObjectScope(Module.COMUtil.PInvokeProxyAPI, 0)
                i += 1
            Module = sys.modules['__main__']
            pid = Module.oDesktop.GetProcessID()
            try:
                os.kill(pid, 9)
                del Module.oDesktop
                return True
            except:
                Module.oMessenger.add_error_message("something went wrong in Closing AEDT")
                return False
        else:
            scopeID = 5
            while i <= scopeID:
                Module.COMUtil.ReleaseCOMObjectScope(Module.COMUtil.PInvokeProxyAPI,i)
                i += 1
            try:
                del Module.oDesktop
            except:
                Module.oMessenger.add_info_message("Attributes not present")
            try:
                del Module.pyaedt_initialized
                return True
            except:
                return False


def force_close_desktop():
    """Close all AEDT projects and forcibly shut down AEDT.

    Returns
    -------
    bool
        ``True`` when successful, ``False`` when failed.

    """
    Module = sys.modules['__main__']
    pid = Module.oDesktop.GetProcessID()
    if pid > 0:
        try:
            plist = Module.oDesktop.GetProjectList()
            for el in plist:
                Module.oDesktop.CloseProject(el)
        except:
            logger.warning("No Projects. Closing Desktop Connection")
        try:
            scopeID = 5
            while i <= scopeID:
                Module.COMUtil.ReleaseCOMObjectScope(Module.COMUtil.PInvokeProxyAPI, 0)
                i += 1
        except:
            logger.warning("No COM UTIL. Closing the Desktop....")
        try:
            del Module.pyaedt_initialized
        except:
            pass
        try:
            os.kill(pid, 9)
            del Module.oDesktop
            successfully_closed = True
        except:
            Module.oMessenger.add_error_message("something went wrong in Closing AEDT")
            successfully_closed = False
        finally:
            log = logging.getLogger(__name__)
            handlers = log.handlers[:]
            for handler in handlers:
                handler.close()
                log.removeHandler(handler)
            return successfully_closed



class Desktop:
    """Initializes AEDT based on the inputs provided.
    
    .. note::
       On Windows, this class works without limitations in IronPython and CPython.
       On Linux, this class works only in embedded IronPython in AEDT.

    Parameters
    ----------
    specified_version : str, optional
        Version of AEDT to use. The default is ``None``, in which case the
        active setup or latest installed version is used.
    NG: bool, optional
        Whether to launch AEDT in the non-graphical mode. The default 
        is ``False``, in which case AEDT launches in the graphical mode.
    AlwaysNew : bool, optional
        Whether to launch an instance of AEDT in a new thread, even if 
        another instance of the ``specified_version`` is active on the machine.
        The default is ``True``.
    release_on_exit : bool, optional
        Whether to release AEDT on exit. The default is ``True``.
    student_version : bool, optional
        Whether to open the AEDT student version. The default is
        ``False``.

    Examples
    --------
    Launch AEDT 2021 R1 in non-graphical mode and initialize HFSS.

    >>> import pyaedt
    >>> desktop = pyaedt.Desktop("2021.1", NG=True)
    pyaedt Info: pyaedt v...
    pyaedt Info: Python version ...
    >>> hfss = pyaedt.Hfss(designname="HFSSDesign1")
    pyaedt Info: Added design 'HFSSDesign1' of type HFSS.

    Launch AEDT 2021 R1 in graphical mode and initialize HFSS.

    >>> desktop = Desktop("2021.1")
    pyaedt Info: pyaedt v...
    pyaedt Info: Python version ...
    >>> hfss = pyaedt.Hfss(designname="HFSSDesign1")
    
    """
            
    @property
    def version_keys(self):
        """Version keys."""

        self._version_keys = []
        self._version_ids = {}
        version_list = list_installed_ansysem()
        for version_env_var in version_list:
            if "ANSYSEMSV_ROOT" in version_env_var:
                current_version_id = version_env_var.replace("ANSYSEMSV_ROOT", '')
                student=True
            else:
                current_version_id = version_env_var.replace("ANSYSEM_ROOT", '')
                student = False
            version = int(current_version_id[0:2])
            release = int(current_version_id[2])
            if version < 20:
                if release < 3:
                    version -= 1
                else:
                    release -= 2
            if student:
                v_key = "20{0}.{1}SV".format(version, release)
                self._version_keys.append(v_key)
                self._version_ids[v_key] = version_env_var
            else:
                v_key = "20{0}.{1}".format(version, release)
                self._version_keys.append(v_key)
                self._version_ids[v_key] = version_env_var

        return self._version_keys

    @property
    def current_version(self):
        """Current version of AEDT."""
        return self.version_keys[0]

    @property
    def current_version_student(self):
        """Current student version of AEDT. """
        for el in self.version_keys:
            if "SV" in el:
                return el
        return None

    def __init__(self, specified_version=None, NG=False, AlwaysNew=True, release_on_exit=True, student_version=False):
        """Initialize desktop."""
        self._main = sys.modules['__main__']
        self._main.close_on_exit = False
        self._main.isoutsideDesktop = False
        self._main.pyaedt_version = pyaedtversion
        self.release = release_on_exit
        self.logfile = None
        module_logger = logging.getLogger(__name__)

        if "oDesktop" in dir(self._main) and self._main.oDesktop is not None:
            self._main.AEDTVersion = self._main.oDesktop.GetVersion()[0:6]
            self._main.oDesktop.RestoreWindow()
            self._main.oMessenger = AEDTMessageManager()
            base_path = self._main.oDesktop.GetExeDir()
            self._main.sDesktopinstallDirectory = base_path
            self.release = False
        else:
            version_student = False
            if specified_version:
                if student_version:
                    specified_version += "SV"
                    version_student = True
                assert specified_version in self.version_keys, \
                    "Specified version {} not known.".format(specified_version)
                version_key = specified_version
            else:
                if student_version and self.current_version_student:
                    version_key = self.current_version_student
                    version_student = True
                else:
                    version_key = self.current_version
                    version_student = False
            base_path = os.getenv(self._version_ids[version_key])
            self._main = sys.modules['__main__']
            self._main.sDesktopinstallDirectory = base_path
            if student_version and version_student:
                version = "Ansoft.ElectronicsDesktop." + version_key[:-2]
            else:
                version = "Ansoft.ElectronicsDesktop." + version_key
            self._main.AEDTVersion = version_key
            self._main.interpreter = _com
            self._main.interpreter_ver = _pythonver
            if "oDesktop" in dir(self._main):
                del self._main.oDesktop
            if _com == 'ironpython':
                sys.path.append(base_path)
                sys.path.append(os.path.join(base_path, 'PythonFiles', 'DesktopPlugin'))
                clr.AddReference("Ansys.Ansoft.CoreCOMScripting")
                AnsoftCOMUtil = __import__("Ansys.Ansoft.CoreCOMScripting")
                self.COMUtil = AnsoftCOMUtil.Ansoft.CoreCOMScripting.Util.COMUtil
                self._main.COMUtil = self.COMUtil
                StandalonePyScriptWrapper = AnsoftCOMUtil.Ansoft.CoreCOMScripting.COM.StandalonePyScriptWrapper
                if NG or AlwaysNew:
                    # forcing new thread to start in non-graphical
                    oAnsoftApp = StandalonePyScriptWrapper.CreateObjectNew(NG)
                else:
                    oAnsoftApp = StandalonePyScriptWrapper.CreateObject(version)
                if NG:
                    os.environ['PYAEDT_DESKTOP_LOGS'] = 'False'
                self._main.oDesktop = oAnsoftApp.GetAppDesktop()
                self._main.isoutsideDesktop = True
            elif _com == 'pythonnet_v3':
                sys.path.append(base_path)
                sys.path.append(os.path.join(base_path, 'PythonFiles', 'DesktopPlugin'))
                launch_msg = "Launching AEDT installation {}".format(base_path)
                print(launch_msg)
                print("===================================================================================")
                clr.AddReference("Ansys.Ansoft.CoreCOMScripting")
                AnsoftCOMUtil = __import__("Ansys.Ansoft.CoreCOMScripting")
                self.COMUtil = AnsoftCOMUtil.Ansoft.CoreCOMScripting.Util.COMUtil
                self._main.COMUtil = self.COMUtil
                StandalonePyScriptWrapper = AnsoftCOMUtil.Ansoft.CoreCOMScripting.COM.StandalonePyScriptWrapper

                module_logger.debug("Launching AEDT with Module Pythonnet")
                processID = []
                if IsWindows:
                    username = getpass.getuser()
                    if student_version:
                        process = "ansysedtsv.exe"
                    else:
                        process = "ansysedt.exe"
                    with os.popen('tasklist /FI "IMAGENAME eq {}" /v'.format(process)) as tasks_list:
                        output = tasks_list.readlines()
                    pattern = r'(?i)^(?:{})\s+?(\d+)\s+.+[\s|\\](?:{})\s+'.format(process, username)
                    for l in output:
                        m = re.search(pattern, l)
                        if m:
                            processID.append(m.group(1))
                if student_version and not processID:
                    import subprocess
                    DETACHED_PROCESS = 0x00000008
                    pid = subprocess.Popen([os.path.join(base_path, "ansysedtsv.exe")],
                                           creationflags=DETACHED_PROCESS).pid
                if NG or AlwaysNew or not processID:
                    # Force new object if no non-graphical instance is running or if there is not an already existing process.
                    App = StandalonePyScriptWrapper.CreateObjectNew(NG)
                else:
                    App = StandalonePyScriptWrapper.CreateObject(version)
                if NG:
                    os.environ['PYAEDT_DESKTOP_LOGS'] = 'False'
                processID2 = []
                if IsWindows:
                    module_logger.debug("Info: Using Windows TaskManager to Load processes")
                    username = getpass.getuser()
                    if student_version:
                        process = "ansysedtsv.exe"
                    else:
                        process = "ansysedt.exe"
                    with os.popen('tasklist /FI "IMAGENAME eq {}" /v'.format(process)) as tasks_list:
                        output = tasks_list.readlines()
                    pattern = r'(?i)^(?:{})\s+?(\d+)\s+.+[\s|\\](?:{})\s+'.format(process, username)
                    for l in output:
                        m = re.search(pattern, l)
                        if m:
                            processID2.append(m.group(1))

                proc = [i for i in processID2 if i not in processID]
                if not proc:
                    if NG:
                        self._main.close_on_exit = False
                    else:
                        self._main.close_on_exit = False
                    oAnsoftApp = win32com.client.Dispatch(version)
                    self._main.oDesktop = oAnsoftApp.GetAppDesktop()
                    self._main.isoutsideDesktop = True
                elif version_key>="2021.1":
                    self._main.close_on_exit = True
                    module_logger.debug("Info: {} Started with Process ID {}".format(version, proc[0]))
                    context = pythoncom.CreateBindCtx(0)
                    running_coms = pythoncom.GetRunningObjectTable()
                    monikiers = running_coms.EnumRunning()
                    for monikier in monikiers:
                        m = re.search(version[10:]+r"\.\d:"+str(proc[0]), monikier.GetDisplayName(context, monikier))
                        if m:
                            obj = running_coms.GetObject(monikier)
                            self._main.isoutsideDesktop = True
                            # self._main.oDesktop = win32com.client.gencache.EnsureDispatch(
                            #     obj.QueryInterface(pythoncom.IID_IDispatch))
                            self._main.oDesktop = win32com.client.Dispatch(obj.QueryInterface(pythoncom.IID_IDispatch))
                            break
                else:
                    module_logger.warning("PyAEDT is not supported in AEDT versions older than 2021.1.")
                    oAnsoftApp = win32com.client.Dispatch(version)
                    self._main.oDesktop = oAnsoftApp.GetAppDesktop()
                    self._main.isoutsideDesktop = True
            else:
                module_logger.debug("Launching AEDT with Module win32com.client")
                oAnsoftApp = win32com.client.Dispatch(version)
                self._main.oDesktop = oAnsoftApp.GetAppDesktop()
                self._main.isoutsideDesktop = True

            self._main.AEDTVersion = version_key
            self._main.oDesktop.RestoreWindow()
            self._main.oMessenger = AEDTMessageManager()
        self._main.pyaedt_initialized = True
        # Set up the log file in the AEDT project directory
        self.logger = logging.getLogger(__name__)
        if not self.logger.handlers:
            project_dir = self._main.oDesktop.GetProjectDirectory()
            self.logfile = os.path.join(project_dir, "pyaedt.log")
            logging.basicConfig(
                filename=self.logfile,
                level=logging.DEBUG,
                format='%(asctime)s:%(name)s:%(levelname)-8s:%(message)s',
                datefmt='%Y/%m/%d %H.%M.%S',
                filemode='w')

        info_msg1 = 'pyaedt v{}'.format(pyaedtversion.strip())
        info_msg2 = 'Python version {}'.format(sys.version)
        self._main.oMessenger.add_info_message(info_msg1, 'Global')
        self._main.oMessenger.add_info_message(info_msg2, 'Global')

        info_msg3 = 'Started external COM connection with module {}'.format(_com)
        info_msg4 = 'Exe path: {}'.format(sys.executable)
        logger.info(info_msg3)
        logger.info(info_msg4)

        if _com == 'pywin32' and (AlwaysNew or NG):
            info_msg5 = 'The ``AlwaysNew`` or ``NG`` option is not available for a pywin32 connection only. Install Python.NET to support these options.'
            self._main.oMessenger.add_info_message(info_msg5, 'Global')
        elif _com == 'ironpython':
            dll_path = os.path.join(base_path,"common","IronPython", "dlls")
            sys.path.append(dll_path)
            info_msg5 = 'Adding IronPython common dlls to the sys.path: {0}'.format(dll_path)
            self._main.oMessenger.add_info_message(info_msg5, 'Global')

    @property
    def install_path(self):
        """Installation path for AEDT."""
        version_key = self._main.AEDTVersion
        root = self._version_ids[version_key]
        return os.environ[root]

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        # Write the trace stack to the log file if an exception occurred in the main script.
        if ex_type:
            err = self._exception(ex_value, ex_traceback)
        if self.release:
            self.release_desktop(close_projects=self._main.close_on_exit, close_on_exit=self._main.close_on_exit)

    def _exception(self, ex_value, tb_data):
        """Write the trace stack to the desktop when a Python error occurs.

        Parameters
        ----------
        ex_value : str
            Type of exception.
        tb_data : str
            Traceback information.
            
        Returns
        -------
        str
            Type of exception.

        """
        try:
            oproject = self._main.oDesktop.GetActiveProject()
            proj_name = oproject.GetName()
            try:
                des_name = oproject.GetActiveDesign().GetName()
                if ";" in des_name:
                    des_name = des_name.split(";")[1]
            except:
                des_name = ''
        except:
            proj_name = ''
            des_name = ''
        tb_trace = traceback.format_tb(tb_data)
        tblist = tb_trace[0].split('\n')
        self._main.oMessenger.add_error_message(str(ex_value), 'Global')
        for el in tblist:
            self._main.oMessenger.add_error_message(el, 'Global')

        return str(ex_value)

    def release_desktop(self, close_projects=True, close_on_exit=True):
        """Release AEDT.

        Parameters
        ----------
        close_projects : bool, optional
            Whether to close the projects opened in the session. 
            The default is ``True``.
        close_on_exit : bool, optional
            Whether to close the active AEDT session on exiting AEDT. 
            The default is ``True``.

        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.release_desktop(close_projects=False, close_on_exit=False) # doctest: +SKIP

        """
        release_desktop(close_projects, close_on_exit)

    def force_close_desktop(self):
        """Close all AEDT projects and shut down AEDT.
    
        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.force_close_desktop() # doctest: +SKIP

        """
        force_close_desktop()

    def close_desktop(self):
        """Close all AEDT projects and shut down AEDT.
    
        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.close_desktop() # doctest: +SKIP

        """
        force_close_desktop()

    def enable_autosave(self):
        """Enable the auto save option.

        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.enable_autosave()

        """
        self._main.oDesktop.EnableAutoSave(True)

    def disable_autosave(self):
        """Disable the auto save option.

        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.disable_autosave()

        """
        self._main.oDesktop.EnableAutoSave(False)


def get_version_env_variable(version_id):
    """Retrieve the environment variable for the AEDT version.

    Parameters
    ----------
    version_id : str
        Full AEDT version number, such as ``"2021.1"``.

    Returns
    -------
    str
        Environment variable for the version.

    Examples
    --------
    >>> from pyaedt import desktop
    >>> desktop.get_version_env_variable("2021.1")
    'ANSYSEM_ROOT211'

    """
    version_env_var = "ANSYSEM_ROOT"
    values = version_id.split('.')
    version = int(values[0][2:])
    release = int(values[1])
    if version < 20:
        if release < 3:
            version += 1
        else:
            release += 2
    version_env_var += str(version) + str(release)
    return version_env_var



"""This module contains the `Simplorer` class."""

from __future__ import absolute_import

import numbers

from .application.AnalysisSimplorer import FieldAnalysisSimplorer
from .application.Variables import Variable
from .desktop import exception_to_desktop
from .generic.general_methods import aedt_exception_handler, generate_unique_name


class Simplorer(FieldAnalysisSimplorer, object):
    """Simplorer class.

    Parameters
    ----------
    projectname : str, optional
        Name of the project to select or the full path to the project
        or AEDTZ archive to open.  The default is ``None``, in which
        case an attempt is made to get an active project. If no 
        projects are present, an empty project is created.
    designname : str, optional
        Name of the design to select. The default is ``None``, in 
        which case an attempt is made to get an active design. If no
        designs are present, an empty design is created.
    solution_type : str, optional
        Solution type to apply to the design. The default is
        ``None``, which applies the default type.
    setup_name : str, optional
        Name of the setup to use as the nominal. The default is
        ``None``, in which case the active setup is used or 
        nothing is used.
    specified_version : str, optional
        Version of AEDT to use. The default is ``None``. If ``None``,
        the active setup is used or the latest installed version is
        used.
    NG : bool, optional
        Whether to launch AEDT in the non-graphical mode. The default
        is ``False``, which launches AEDT in the graphical mode.
    AlwaysNew : bool, optional
        Whether to launch an instance of AEDT in a new thread, even if
        another instance of the ``specified_version`` is active on the
        machine.  The default is ``True``.
    release_on_exit : bool, optional
        Whether to release AEDT on exit. The default is ``True``.
    student_version : bool, optional
        Whether open AEDT Student Version. The default is ``False``.

    Examples
    --------
    Create an instance of `Simplorer` and connect to an existing
    Maxwell design or create a new Maxwell design if one does not
    exist.

    >>> from pyaedt import Simplorer
    >>> app = Simplorer()

    Create a instance of `Simplorer` and link to a project named
    ``"projectname"``. If this project does not exist, create one with
    this name.

    >>> app = Simplorer(projectname)

    Create an instance of `Simplorer` and link to a design named
    ``"designname"`` in a project named ``"projectname"``.

    >>> app = Simplorer(projectname, designame)

    Create an instance of `Simplorer` and open the specified
    project, which is named ``"myfile.aedt"``.

    >>> app = Simplorer("myfile.aedt")
    """

    def __init__(self, projectname=None, designname=None, solution_type=None, setup_name=None,
                 specified_version=None, NG=False, AlwaysNew=False, release_on_exit=False, student_version=False):
        """Constructor."""
        FieldAnalysisSimplorer.__init__(self, "Twin Builder", projectname, designname, solution_type, setup_name,
                                        specified_version, NG, AlwaysNew, release_on_exit,student_version)


    @aedt_exception_handler
    def create_schematic_from_netlist(self, file_to_import):
        """Create a circuit schematic from an HSpice net list.
        
        Supported currently:
        
        * R
        * L
        * C
        * Diodes
        * Bjts
        * Discrete components with syntax ``Uxxx net1 net2 ... netn modname``

        Parameters
        ----------
        file_to_import : str
            Full path to the HSpice file.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        xpos = 0
        ypos = 0
        delta = 0.0508
        use_instance = True
        with open(file_to_import, 'r') as f:
            for line in f:
                mycomp = None
                fields = line.split(" ")
                name = fields[0]
                if fields[0][0] == "R":
                    value = fields[3][fields[3].find("=") + 1:].strip()
                    mycomp, mycompname = self.modeler.components.create_resistor(name, value, xpos, ypos,
                                                                                 use_instance_id_netlist=use_instance)
                elif fields[0][0] == "L":
                    value = fields[3][fields[3].find("=") + 1:].strip()
                    mycomp, mycompname = self.modeler.components.create_inductor(name, value, xpos, ypos,
                                                                                 use_instance_id_netlist=use_instance)
                elif fields[0][0] == "C":
                    value = fields[3][fields[3].find("=") + 1:].strip()
                    mycomp, mycompname = self.modeler.components.create_capacitor(name, value, xpos, ypos,
                                                                                  use_instance_id_netlist=use_instance)
                elif fields[0][0] == "Q":
                    if len(fields) == 4 and fields[0][0] == "Q":
                        value = fields[3].strip()
                        mycomp, mycompname = self.modeler.components.create_npn(fields[0], value, xpos, ypos,
                                                                               use_instance_id_netlist=use_instance)
                        value = None
                elif fields[0][0] == "D":
                    value = fields[3][fields[3].find("=") + 1:].strip()
                    mycomp, mycompname = self.modeler.components.create_diode(name, value, xpos, ypos,
                                                                              use_instance_id_netlist=use_instance)
                if mycomp:
                    pins = self.modeler.components.get_pins(mycomp)
                    id = 1
                    for pin in pins:
                        if pin == "CH" or pin==fields[0][0]:
                            continue
                        pos = self.modeler.components.get_pin_location(mycomp, pin)
                        if pos[0] < xpos:
                            angle = 6.28318530717959
                        else:
                            angle = 3.14159265358979
                        self.modeler.components.create_page_port(fields[id], pos[0], pos[1], angle)
                        id += 1
                    ypos += delta
                    if ypos > 0.254:
                        xpos += delta
                        ypos = 0
        return True


    @aedt_exception_handler
    def set_end_time(self, expression):
        """

        Parameters
        ----------
        expression :
            

        Returns
        -------

        """
        self.set_sim_setup_parameter('Tend', expression)
        return True

    @aedt_exception_handler
    def set_hmin(self, expression):
        """

        Parameters
        ----------
        expression :
            

        Returns
        -------

        """
        self.set_sim_setup_parameter('Hmin', expression)
        return True

    @aedt_exception_handler
    def set_hmax(self, expression):
        """

        Parameters
        ----------
        expression :
            

        Returns
        -------

        """
        self.set_sim_setup_parameter('Hmax', expression)
        return True

    @aedt_exception_handler
    def set_sim_setup_parameter(self, var_str, expression, analysis_name="TR"):
        """

        Parameters
        ----------
        var_str : string
            Name of the variable.
            
        expression :
            
        analysis_name : str, optional
             Name of the analysis. The default is ``"TR"``.

        Returns
        -------

        """
        if isinstance(expression, Variable):
            value_str = expression.string_value
        # Handle input type int/float, etc (including numeric 0)
        elif isinstance(expression, numbers.Number):
            value_str = str(expression)
        else:
            value_str = expression

        self._odesign.ChangeProperty(
            [
                "NAME:AllTabs",
                [
                    "NAME:BaseElementTab",
                    [
                        "NAME:PropServers",
                        analysis_name
                    ],
                    [
                        "NAME:ChangedProps",
                        [
                            "NAME:" + var_str,
                            "Value:="	, value_str
                        ]
                    ]
                ]
            ])
        return True

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        """ Push exit up to parent object Design """
        if ex_type:
            exception_to_desktop(self, ex_value, ex_traceback)
